# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import errno
import json
import requests
import sentry_sdk

# noinspection PyPackageRequirements
import flask
import serial
import tornado

from octoprint.events import Events
import octoprint.plugin
import octoprint.settings
from octoprint.util.commandline import CommandlineError

from octoprint_simplyprint.websocket import SimplyPrintWebsocket
from octoprint_simplyprint.comm import SimplyPrintComm
from octoprint_simplyprint.local import cron

SIMPLYPRINT_EVENTS = [
    Events.PRINTER_STATE_CHANGED,
    Events.TOOL_CHANGE,
    Events.CONNECTING,
    Events.CONNECTED,
    Events.DISCONNECTING,
    Events.DISCONNECTED,

    Events.STARTUP,
    Events.SHUTDOWN,

    Events.ERROR,

    Events.FILE_SELECTED,

    Events.PRINT_STARTED,
    Events.PRINT_FAILED,
    Events.PRINT_DONE,
    Events.PRINT_CANCELLING,
    Events.PRINT_CANCELLED,
    Events.PRINT_PAUSED,
    Events.PRINT_RESUMED,

    "plugin_printer_safety_check_warning",
    Events.FIRMWARE_DATA,

    "plugin_bedlevelvisualizer_mesh_data_collected",

    "plugin_pluginmanager_install_plugin",
    "plugin_pluginmanager_uninstall_plugin",
    "plugin_pluginmanager_enable_plugin",
    "plugin_pluginmanager_disabled_plugin",

    "plugin_simplyfilamentsensor_filament_loaded",
    "plugin_simplyfilamentsensor_filament_runout",
    "plugin_simplyfilamentsensor_filament_no_filament_print_on_print_start",
    "plugin_simplyfilamentsensor_no_filament_on_print_start_paused",
    "plugin_simplyfilamentsensor_no_filament_on_print_start_cancelled",

    "plugin_psucontrol_psu_state_changed",

    "plugin_simplypowercontroller_power_on",
    "plugin_simplypowercontroller_power_off",

    Events.METADATA_ANALYSIS_FINISHED,

    "plugin_firmware_check_warning",
    Events.FILE_REMOVED,
]

IGNORED_EXCEPTIONS = [
    # serial exceptions in octoprint.util.comm
    (
        serial.SerialException,
        lambda exc, logger, plugin, cb: logger == "octoprint.util.comm",
    ),
    # KeyboardInterrupts
    KeyboardInterrupt,
    # IOErrors of any kind due to a full file system
    (
        IOError,
        lambda exc, logger, plugin, cb: exc.errorgetattr(exc, "errno")  # noqa: B009
        and exc.errno in (getattr(errno, "ENOSPC"),),  # noqa: B009
    ),
    # RequestExceptions of any kind
    requests.exceptions.RequestException,
    # Tornado WebSocketErrors of any kind
    tornado.websocket.WebSocketError,
    # Tornado HTTPClientError
    tornado.httpclient.HTTPClientError,
    # error from windows for linux specific commands related to wifi
    CommandlineError
]

class SimplyPrint(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.ShutdownPlugin,
    octoprint.plugin.BlueprintPlugin,
):
    _files_analyzed = []

    simply_print = None
    port = "5000"

    def initialize(self):
        # Called once the plugin has been loaded by OctoPrint, all injections complete
        sp_cls = SimplyPrintComm
        if self._settings.get_boolean(["websocket_ready"]):
            sp_cls = SimplyPrintWebsocket
        # elif not self._settings.get_boolean(["is_set_up"]):
        #     # If not setup default new accounts to the websocket
        #     sp_cls = SimplyPrintWebsocket
        #     self._settings.set_boolean(["websocket_ready"], True)
        #     self._settings.save()
        self.simply_print = sp_cls(self)

    def on_startup(self, host, port):
        # Initialize sentry.io for error tracking
        self._initialize_sentry()
        # Run startup thread and run the main loop in the background
        if isinstance(self.simply_print, SimplyPrintComm):
            self.simply_print.start_startup()
            self.simply_print.start_main_loop()
        else:
            self.simply_print.on_startup()

        # Remember that this port is internal to OctoPrint, a proxy may exist.
        self.port = port
        if port != 5000 and port != 80 and port != 443:
            self.send_port_ip(port)

    def notify_websocket_ready(self):
        cron.remove_cron_jobs()
        self._settings.set_boolean(["websocket_ready"], True)
        self._settings.save()
        self.simply_print = SimplyPrintWebsocket(self)
        self.simply_print.on_startup()

    # #~~ StartupPlugin mixin
    def on_after_startup(self):

        self._logger.info("SimplyPrint OctoPrint plugin started")

        # If cron jobs don't exist, create them
        if (
            isinstance(self.simply_print, SimplyPrintComm)
            and not cron.check_cron_jobs()
        ):
            cron.create_cron_jobs()

        # The "Startup" event is never picked up by the plugin, as the plugin is loaded AFTER startup
        self.on_event("Startup", {})

    def _initialize_sentry(self):
        self._logger.debug("Initializing Sentry")

        def _before_send(event, hint):
            if "exc_info" not in hint:
                # we only want exceptions
                return None

            handled = True
            logger = event.get("logger", "")
            plugin = event.get("extra", {}).get("plugin", None)
            callback = event.get("extra", {}).get("callback", None)

            for ignore in IGNORED_EXCEPTIONS:
                if isinstance(ignore, tuple):
                    ignored_exc, matcher = ignore
                else:
                    ignored_exc = ignore
                    matcher = lambda *args: True

                exc = hint["exc_info"][1]
                if isinstance(exc, ignored_exc) and matcher(
                    exc, logger, plugin, callback
                ):
                    # exception ignored for logger, plugin and/or callback
                    return None

                elif isinstance(ignore, type):
                    if isinstance(hint["exc_info"][1], ignore):
                        # exception ignored
                        return None

            # if event.get("exception") and event["exception"].get("values"):
            #     handled = not any(
            #         map(
            #             lambda x: x.get("mechanism")
            #             and not x["mechanism"].get("handled", True),
            #             event["exception"]["values"],
            #         )
            #     )
            #
            # if handled:
            #     # error is handled, restrict further based on logger
            #     if logger != "" and not (
            #         logger.startswith("octoprint.plugins.SimplyPrint") or logger.startswith("octoprint.plugins.simplyprint")
            #     ):
            #         # we only want errors logged by our plugin's loggers
            #         return None

            if logger.startswith("octoprint.plugins.SimplyPrint") or logger.startswith("octoprint.plugins.simplyprint"):
                return event
            else:
                return None

        sentry_sdk.init(
            dsn="https://c35fae8df2d74707bec50279a0bcd7ae@o1102514.ingest.sentry.io/6611344",
            traces_sample_rate=0.01,
            before_send=_before_send,
            release="SimplyPrint@{}".format(self._plugin_version)
        )
        if self._settings.get(["printer_id"]) != "":
            sentry_sdk.set_user({"id": self._settings.get(["printer_id"])})

    def on_shutdown(self):
        if isinstance(self.simply_print, SimplyPrintComm):
            # SimplyPrintComm will stop on next loop
            self.simply_print.run_loop = False
        elif self.simply_print is not None:
            self.simply_print.close()

    @staticmethod
    def get_settings_defaults():
        return {
            "request_url": "",
            "rpi_id": "",
            "is_set_up": False,
            "printer_name": "",
            "printer_id": "",
            "temp_short_setup_id": "",
            "from_image": False,
            "sp_installed_plugins": [],
            "display_enabled": True,
            "display_branding": True,
            "display_show_status": True,
            "display_while_printing_type": "0",
            "has_power_controller": False,
            "has_filament_sensor": False,
            "webcam": {
                "flipH": False,
                "flipV": False,
                "rotate90": False
            },
            "info": {
                "last_user_settings_sync": "0000-00-00 00:00:00",
                "gcode_scripts_backed_up": False,
            },
            "debug_logging": False,
            "public_port": "80",
            # Websocket Default Settings
            "websocket_ready": True,
            "endpoint": "production",
            "printer_token": "",
            "ambient_temp": "85",
        }

    def on_settings_save(self, data):
        octoprint.plugin.SettingsPlugin.on_settings_save(self, data)
        new_printer_id = self._settings.get(["printer_id"])

        if new_printer_id != "":
            sentry_sdk.set_user({"id": self._settings.get(["printer_id"])})

    def get_template_vars(self):
        return {
            "version": self._plugin_version
        }

    @staticmethod
    def get_assets():
        return dict(
            js=["js/SimplyPrint.js"],
            css=["css/SimplyPrint.css"],
            font=["font/lcd.ttf"],
            logo=["img/sp_logo.png"],
            logo_lg=["img/sp_logo_large.png"],
            logo_white_sm=["img/sp_white_sm.png"]
        )

    def get_api_commands(self):
        return {
            "setup": [],  # Sets up SimplyPrintRPiSoftware
            "uninstall": [],  # Uninstalls SimplyPrintRPiSoftware
            "message": ["payload"], # Inject websocket messages
        }

    @staticmethod
    def _uninstall_sp():
        # All we need to do here is remove cron jobs, nothing else
        cron.remove_cron_jobs()

    @staticmethod
    def _install_background():
        if not cron.check_cron_jobs():
            cron.create_cron_jobs()

    def on_api_command(self, command, data):
        if isinstance(self.simply_print, SimplyPrintWebsocket):
            if command == "message" and self.simply_print.test:
                msg = json.dumps(data["payload"])
                # Generally we do NOT want to access methods marked
                # as private, however this is for testing only
                self.simply_print._process_message(msg)
            return
        if command == "setup":
            self._uninstall_sp()
        elif command == "uninstall":
            self._uninstall_sp()

    # Send public port to outside system
    def send_port_ip(self, port=None):
        self._settings.set(["public_port"], port)
        self._settings.save()

    def on_api_get(self, request):
        import flask
        import subprocess
        # self.log(str(request))
        # self.log(str(request.args))

        if request.args is not None:
            if request.args.get("install", default=None, type=None) is not None:
                # Install
                pass
            if request.args.get("send_port", default=None, type=None) is not None:
                # Send port to local scripts
                port = str(request.args.get("send_port", default=None, type=None))
                self.send_port_ip(port)
            if request.args.get("rpi_id", default=None, type=None) is not None:
                # Get RPI id
                pass
            if request.args.get("do_gcode", default=None, type=None) is not None:
                # Execute GCODE
                gcode_todo = str(request.args.get("do_gcode", default=None, type=None))
                self._printer.commands(gcode_todo.split(","))
                pass
            if request.args.get("power_controller", default=None, type=None) is not None:
                # Power Controller
                new_state = str(request.args.get("power_controller", default=None, type=None))
                psu_state = ""
                if new_state == "1":
                    # Turn Power Controller on
                    psu_state = "turnPSUOn"
                elif new_state == "0":
                    # Turn Power Controller off
                    psu_state = "turnPSUOff"
                elif new_state == "get":
                    # Turn Power Controller off
                    psu_state = "getPSUState"

                try:
                    r = requests.post("http://localhost/api/plugin/psucontrol", data={"command": psu_state},
                                      allow_redirects=True, verify=False)
                    r.raise_for_status()

                    # Parse
                    try:
                        the_json = json.loads(r.content)
                    except:
                        self._logger.error("Failed to format request response to JSON; " + str(r.content))
                        return False
                except:
                    pass

    # Blueprint mixin
    @octoprint.plugin.BlueprintPlugin.route("/can_reboot", methods=["GET"])
    def can_reboot_route(self):
        can_reboot = self._printer.get_state_id() not in ["PRINTING", "PAUSED", "PAUSING"]
        self._logger.debug("Is it ok to reboot? {}".format(can_reboot))
        return flask.jsonify({"can_reboot": can_reboot})

    def is_blueprint_protected(self):
        return False

    # EventHandler mixin
    def on_event(self, event, payload):
        if event in SIMPLYPRINT_EVENTS:
            self.simply_print.on_event(event, payload)

    def gcode_sent(self, comm_instance, phase, cmd, cmd_type, gcode, *args, **kwargs):
        # if gcode and gcode == "M106":
        #     self._logger.info("Just sent M106: {cmd}".format(**locals()))
        tags = kwargs.get("tags", [])
        if tags is None:
            return
        if (
            ("source:api" in tags or "plugin:octoprint_simplyprint" in tags) and
            isinstance(self.simply_print, SimplyPrintWebsocket)
        ):
            self.simply_print.on_gcode_sent(cmd)

    def gcode_received(self, comm_instance, line, *args, **kwargs):
        if line.strip() not in ["echo:busy: paused for user", "echo:busy: processing", "Unknown M code: M118 simplyprint unpause", "simplyprint unpause"]:
            return line

        if line.strip() == "echo:busy: paused for user":
            self._logger.debug("received line: echo:busy: paused for user, setting user_input_required True")
            self.simply_print.user_input_required = True
            self._printer.commands("M118 simplyprint unpause", force=True)
        if self.simply_print.user_input_required and line.strip() in ["echo:busy: processing", "Unknown M code: M118 simplyprint unpause", "simplyprint unpause"]:
            self._logger.debug("received line: echo:busy: processing, setting user_input_required False")
            self.simply_print.user_input_required = False

        if isinstance(self.simply_print, SimplyPrintWebsocket):
            self.simply_print.on_gcode_received(line)

        return line

    def process_at_command(self, comm, phase, command, parameters, tags=None, *args, **kwargs):
        if command.lower() not in ["simplyprint", "pause"]:
            return

        if isinstance(self.simply_print, SimplyPrintWebsocket):
            cmd = command.lower()
            if cmd == "pause":
                self.simply_print.on_pause_at_command(parameters)
            elif cmd == "simplyprint":
                params = parameters.strip().split(" ")
                if params and params[0] ==  "layer":
                    try:
                        layer = int(params[1])
                    except Exception:
                        pass
                    else:
                        self.simply_print.on_layer_change(layer)
            return

        url_parameters = ""
        if command.lower() == "pause":
            url_parameters += "&pause_message={}".format(parameters)
        elif command.lower() == "simplyprint":
            if parameters:
                parameters_list = parameters.split(" ")
            if parameters_list[0] == "layer":
                url_parameters += "&layer={}".format(parameters_list[1])

        if url_parameters != "":
            self.simply_print.ping(url_parameters)

        return

    def get_update_information(self):
        return dict(
            SimplyPrint=dict(
                displayName="SimplyPrint",
                displayVersion=self._plugin_version,

                # version check: github repository
                type="github_release",
                user="SimplyPrint",
                repo="OctoPrint-SimplyPrint",
                current=self._plugin_version,
                stable_branch=dict(name="Stable", branch="master", comittish=["master"]),
                prerelease_branches=[
                    dict(
                        name="Development",
                        branch="devel",
                        comittish=["develop", "rc", "master"],
                    ),
                    dict(
                        name="Release Candidate",
                        branch="rc",
                        comittish=["rc", "master"],
                    )
                ],
                # update method: pip
                pip="https://github.com/SimplyPrint/OctoPrint-SimplyPrint/archive/{target_version}.zip"
            )
        )


__plugin_name__ = "SimplyPrint Cloud"
__plugin_pythoncompat__ = ">=3.7,<4"
__plugin_disabling_discouraged__ = """
Please uninstall SimplyPrint Cloud rather than just disable it, since it sets up some background scripts
that will continue to run if you disable it.
"""
# Remember to bump the version in setup.py as well
__plugin_version__ = "4.0.6"

def __plugin_load__():
    global __plugin_implementation__, __plugin_hooks__
    __plugin_implementation__ = SimplyPrint()
    __plugin_hooks__ = {
        "octoprint.plugin.softwareupdate.check_config": __plugin_implementation__.get_update_information,
        "octoprint.comm.protocol.atcommand.sending": __plugin_implementation__.process_at_command,
        "octoprint.comm.protocol.gcode.received": __plugin_implementation__.gcode_received,
        "octoprint.comm.protocol.gcode.sent": __plugin_implementation__.gcode_sent
    }



# -*- coding: utf-8 -*-
#
# SimplyPrint
# Copyright (C) 2020-2022  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations
import os
import logging
import socket
import sys
import io
import re
import ipaddress
import json
import psutil
import requests
import sarge
import platform
from typing import TYPE_CHECKING, Callable, List, Dict, Any, Optional, Tuple

from octoprint.util.commandline import CommandlineCaller
from octoprint.util.pip import LocalPipCaller
from octoprint.plugin import PluginSettings
from octoprint.util.platform import CLOSE_FDS
from octoprint.events import Events, EventManager
from .constants import *

if TYPE_CHECKING:
    from .simplyprint import SimplyPrintWebsocket

class SystemQuery:
    def __init__(self, settings: PluginSettings) -> None:
        self._settings = settings
        self._logger = logging.getLogger("octoprint.plugins.simplyprint")
        self.command_line = CommandlineCaller()

    def get_system_info(self) -> Dict[str, Any]:
        info = {}
        ver, api_ver = self._get_octoprint_version()
        info["ui_version"] = ver
        info["api_version"] = api_ver
        info["python_version"] = self._get_python_version()
        info["machine"] = self._get_cpu_model()
        info["os"] = platform.system()
        info["core_count"] = os.cpu_count()
        info["total_memory"] = psutil.virtual_memory().total
        info.update(self.get_network_info())

        port = self._get_public_port()

        if port:
            info["local_ip"] += ":" + str(port)

        return info

    def _get_python_version(self) -> str:
        return ".".join(str(part) for part in sys.version_info[:3])

    def _get_routed_ip(self) -> Optional[str]:
        src_ip = None
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.settimeout(0)
            s.connect(('10.255.255.255', 1))
            src_ip = s.getsockname()[0]
        except Exception:
            pass
        finally:
            s.close()
        if src_ip == "" or src_ip == "127.0.1.1":
            src_ip = None
        return src_ip

    def _get_wifi_interface(self) -> Tuple[Optional[str], Optional[str]]:
        os_name = platform.system()
        try:
            if os_name == "Linux":
                cmd = "iwgetid"
                ret, stdout, stderr = self.command_line.checked_call(cmd)
                if stdout:
                    parts = stdout[0].strip().split(maxsplit=1)
                    ssid = parts[1].split(":")[-1].strip('"')
                    return parts[0], ssid
            elif os_name == "Windows":
                cmd = ["netsh", "wlan", "show", "interfaces"]
                ret, stdout, stderr = self.command_line.checked_call(cmd)
                if stdout:
                    for line in stdout:
                        if " SSID" in line:
                            return None, line.split(":")[1].strip()

        except Exception:
            self._logger.exception("Failed to retreive wifi interfaces")
        return None, None

    def get_network_info(self) -> Dict[str, Any]:
        src_ip = self._get_routed_ip()
        netinfo = {
            "hostname": socket.gethostname(),
            "local_ip": src_ip or "",
            "ssid": "",
            "is_ethernet": True
        }
        wifi_intf, ssid = self._get_wifi_interface()
        cmd = "ip -json address"
        try:
            ret, stdout, stderr = self.command_line.checked_call(cmd)
            decoded = json.loads("\n".join(stdout))
            for interface in decoded:
                if (
                    interface['operstate'] != "UP" or
                    interface['link_type'] != "ether" or
                    'address' not in interface
                ):
                    continue
                ifname = interface["ifname"]
                for addr in interface.get('addr_info', []):
                    if "family" not in addr or "local" not in addr:
                        continue
                    if addr["family"] == "inet6" and src_ip is None:
                        ip = ipaddress.ip_address(addr["local"])
                        if ip.is_global:
                            netinfo["local_ip"] = addr["local"]
                            if wifi_intf == ifname:
                                netinfo["is_ethernet"] = False
                                netinfo["ssid"] = ssid
                            return netinfo
                    elif src_ip == addr["local"]:
                        netinfo["local_ip"] = addr["local"]
                        if wifi_intf == ifname:
                            netinfo["is_ethernet"] = False
                            netinfo["ssid"] = ssid
                        return netinfo
        except Exception:
            self._logger.exception("Failed to parse network interfaces")
        return netinfo

    def _get_cpu_model(self) -> str:
        info_path = "/proc/cpuinfo"
        try:
            with io.open(info_path, "rt", encoding="utf-8") as file:
                data = file.read()
            cpu_items = [
                item.strip() for item in data.split("\n\n") if item.strip()
            ]
            # Check for the Raspberry Pi Model
            match = re.search(r"Model\s+:\s+(.+)", cpu_items[-1])
            if match is not None:
                return match.group(1).strip()
            # Check for the CPU model typically reported by desktop
            # class machines
            for item in cpu_items:
                match = re.search(r"model name\s+:\s+(.+)", item)
                if match is not None:
                    return match.group(1).strip()
        except Exception:
            pass
        return "unknown"

    def _get_octoprint_version(self) -> Tuple[str, str]:
        """
        Get OctoPrint version and API version
        :return: (tuple) OctoPrint version, API version
        """
        from octoprint.server.api import VERSION
        from octoprint import __version__
        return __version__, VERSION

    def _get_public_port(self) -> str:
        # noinspection PyProtectedMember
        return self._settings.get(["public_port"])


class SystemManager:
    def __init__(self, simplyprint: SimplyPrintWebsocket) -> None:
        self.simplyprint = simplyprint
        self.logger = self.simplyprint._logger
        self.settings = self.simplyprint.settings
        self.installed_plugins: List[str] = self.settings.get(["sp_installed_plugins"])
        if not isinstance(self.installed_plugins, list):
            self.installed_plugins = []
            self.settings.set(["sp_installed_plugins"], [])
            self.settings.save()
        self._pip_caller = LocalPipCaller(
            # Use --user on commands if user has configured it
            force_user=self.settings.global_get_boolean(
                ["plugins", "pluginmanager", "pip_force-user"]
            )
        )

    def update_simplyprint(self) -> None:
        test = self.simplyprint.test
        url = TEST_PLUGIN_INSTALL_URL if test else PLUGIN_INSTALL_URL
        data = {
            "name": "SimplyPrint",
            "pip_name": "SimplyPrint",
            "key": "SimplyPrint",
            "install_url": url
        }
        if self.install_plugin(data):
            self.restart_octoprint()

    def install_plugin(self, plugins: List[Dict[str, str]]) -> bool:
        install_error = False
        for plugin_data in plugins:
            name = plugin_data['name']
            url = plugin_data["install_url"]
            self.logger.info(
                f"Installing OctoPrint Plugin '{name}' at "
                f"request from SimplyPrint from {url}"
            )
            key = plugin_data.get("key")
            if key is not None and key not in self.installed_plugins:
                self.installed_plugins.append(key)
                self.settings.set(["sp_installed_plugins"], self.installed_plugins)
                self.settings.save()
            args: List[str] = ["install", url, "--no-cache-dir"]
            try:
                code, stdout, stderr = self._call_pip(*args)
            except Exception:
                self.logger.exception(f"Failed to install plugin: {name}")
                install_error = True
            if code != 0:
                self.logger.error(
                    f"Failed to install plugin {name}, returned with {code}\n"
                    f"{stdout}\n{stderr}"
                )
                install_error = True
        if install_error:
            return False
        return True

    def uninstall_plugin(self, plugin_data: Dict[str, str]) -> bool:
        name = plugin_data['name']
        self.logger.info(
            f"Uninstalling OctoPrint Plugin '{name}' at "
            "request from SimplyPrint"
        )
        key = plugin_data.get("key")
        if key is not None and key in self.installed_plugins:
            self.installed_plugins.remove(key)
            self.settings.set(["sp_installed_plugins"], self.installed_plugins)
            self.settings.save()
        pip_name = plugin_data["pip_name"].replace(" ", "-")
        args: List[str] = ["uninstall", "--yes", pip_name]
        try:
            code, stdout, stderr = self._call_pip(*args)
        except Exception:
            self.logger.exception(f"Failed to uninstall plugin {name}")
            return False
        if code != 0:
            self.logger.error(
                f"Failed to uninstall plugin {name}, returned with {code}\n"
                f"{stdout}\n{stderr}"
            )
            return False
        return True

    def restart_octoprint(self) -> None:
        self.logger.info("Restarting OctoPrint at request from SimplyPrint")
        command = self.settings.global_get(
            ["server", "commands", "serverRestartCommand"]
        )
        if command:
            self.simplyprint.close()
            self.simplyprint.event_bus.fire(Events.SHUTDOWN)
            self._run_system_command(command, do_async=True)

    def stop_octoprint(self) -> None:
        self.logger.info("Stopping OctoPrint at request from SimplyPrint")
        restart_cmd = self.settings.global_get(
            ["server", "commands", "serverRestartCommand"]
        )
        command = ""
        if isinstance(restart_cmd, str):
            restart_cmd = restart_cmd.strip()
            if restart_cmd.startswith("sudo service "):
                svc = restart_cmd.split()[2]
                command = f"sudo service {svc} stop"
            elif command.startswith("sudo systemctl"):
                svc = restart_cmd.split()[-1]
                command = f"sudo systemctl stop {svc}"
        if command:
            self.simplyprint.close()
            self.simplyprint.event_bus.fire(Events.SHUTDOWN)
            self._run_system_command(command, do_async=True)

    def shutdown_machine(self) -> None:
        self.logger.info("Machine shutdown at request from SimplyPrint")
        command = self.settings.global_get(
            ["server", "commands", "systemShudownCommand"]
        )
        if command:
            self.simplyprint.close()
            self.simplyprint.event_bus.fire(Events.SHUTDOWN)
            self._run_system_command(command, do_async=True)

    def reboot_machine(self) -> None:
        self.logger.info("Machine reboot at request from SimplyPrint")
        command = self.settings.global_get(
            ["server", "commands", "systemRestartCommand"]
        )
        if command:
            self.simplyprint.close()
            self.simplyprint.event_bus.fire(Events.SHUTDOWN)
            self._run_system_command(command, do_async=True)

    def _call_pip(self, *args) -> Any:
        if self._pip_caller is None or not self._pip_caller.available:
            self.logger.error("No pip available, can't install plugin")
            raise Exception("No pip available, can't install")
        return self._pip_caller.execute(*args)

    def power_on_printer(self) -> None:
        self._do_power_action("on")

    def power_off_printer(self) -> None:
        self._do_power_action("off")

    def get_power_state(self) -> Optional[bool]:
        if not self.settings.get_boolean(["has_power_controller"]):
            return None
        helpers = self.simplyprint.get_helpers("psucontrol")
        if helpers is None:
            helpers = self.simplyprint.get_helpers("simplypowercontroller")
            if helpers is not None and "get_status" in helpers:
                status = helpers["get_status"]()
                return status["isPSUOn"]

        elif "get_psu_state" in helpers:
            return helpers["get_psu_state"]()
        return None

    def _do_power_action(self, action: str) -> None:
        helpers = self.simplyprint.get_helpers("psucontrol")
        func: Optional[Callable] = None
        if helpers is None:
            helpers = self.simplyprint.get_helpers("simplypowercontroller")
            if helpers is not None:
                func = helpers.get(f"psu_{action}")
        else:
            # psu control is available
            func = helpers.get(f"turn_psu_{action}")
        if func is not None:
            func()

    def get_filament_sensor_state(self) -> Optional[str]:
        if not self.settings.get_boolean(["has_filament_sensor"]):
            return None
        helpers = self.simplyprint.get_helpers("simplyfilamentsensor")
        if helpers is not None and "get_status" in helpers:
            status = helpers["get_status"]()
            return "loaded" if status["has_filament"] else "runout"
        return None

    def _run_system_command(
        self, command: str, do_async: bool = False
    ) -> None:
        try:
            ret = sarge.run(
                command, close_fds=CLOSE_FDS, stdout=sarge.Capture(),
                stderr=sarge.Capture(), async_=do_async
            )
        except Exception:
            self.logger.exception(f"Error running command: {command}")
            return
        if not do_async and ret.returncode != 0:
            stdout = ret.stdout.text  # type: ignore
            stderr = ret.stderr.text  # type: ignore
            self.logger.error(
                f"Failed to run command '{command}', returned with "
                f"{ret.returncode}\n{stdout}\n{stderr}"
            )

    def check_software_update(self) -> List[Dict[str, Any]]:
        port = self.simplyprint.plugin.port
        api_key = self.settings.global_get(["api", "key"])
        url = f"http://127.0.0.1:{port}/plugin/softwareupdate/check"
        try:
            resp = requests.get(
                url, headers={"X-Api-Key": api_key}, timeout=5
            )
            if not 200 <= resp.status_code <= 210:
                # Response code no good
                self.logger.warning("Couldn't check for an OctoPrint update, API returned invalid response")
                return []
            ret: Dict[str, Any] = resp.json()
        except Exception:
            self.logger.warning("Error fetching OctoPrint Updates")
            return []
        updates: List[Dict[str, Any]] = []
        uinfo: Dict[str, Any]
        for name, uinfo in ret.get("information", {}).items():
            if uinfo.get("updateAvailable", False):
                local_ver = uinfo["information"]["local"]["value"]
                remote_ver = uinfo["information"]["remote"]["value"]
                updates.append({
                    "plugin": name,
                    "local_version": local_ver,
                    "available_version": remote_ver,
                    "release_notes": uinfo["releaseNotes"]
                })
        return updates



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import threading
import time
import uuid

import io
import os
import tempfile
try:
    # Python 3
    from queue import Queue
except ImportError:
    # Python 2
    from Queue import Queue


import requests

# noinspection PyPackageRequirements
from octoprint.settings import settings

from .constants import WEBCAM_SNAPSHOT_URL

log = logging.getLogger("octoprint.plugins.SimplyPrint.comm.webcam")


def post_image(picture_id=None):
    log.debug("Taking picture")

    if picture_id is None:
        picture_id = str(uuid.uuid1())
        upl_url = WEBCAM_SNAPSHOT_URL + "?livestream=" + settings().get(["plugins", "SimplyPrint", "rpi_id"])
    else:
        upl_url = WEBCAM_SNAPSHOT_URL + "?request_id=" + picture_id

    if settings().get(["webcam", "webcamEnabled"]):
        url = settings().get(["webcam", "snapshot"])

        if url is None:
            log.warning("Webcam URL is None")
            return

        temp_dir = tempfile.gettempdir()
        temp_path = os.path.join(temp_dir, "simplyprint-webcam-{}.png".format(uuid.uuid1()))

        with io.open(temp_path, "wb") as file:
            log.debug("Downloading image from {}".format(url))
            try:
                download_image(file, url)
            except WebcamError:
                requests.get(upl_url + "&err_msg=Download%20of%20screenshot%20not%20successful", timeout=5)
                return

        with io.open(temp_path, "rb") as file:
            try:
                response = requests.post(upl_url, files={"the_file": file}, timeout=10)
            except requests.exceptions.RequestException as e:
                log.error("Failed to post image to SP")
                log.error(repr(e))
                return

            if int(round(response.status_code / 100)) != 2:
                log.error("Received non-2xx status code from SP for url {}, content: {}".format(upl_url,
                                                                                                response.content))

        try:
            os.remove(temp_path)
        # except FileNotFoundError:
        #    pass
        except Exception:
            log.warning("Could not remove webcam snapshot tempfile")

        return response


def download_image(file_object, url, timeout=3):
    if settings().get(["plugins", "SimplyPrint", "debug_logging"]):
        log.setLevel(logging.DEBUG)
    """
    Download an image from the webcam and write it to the file object
    :param file_object: File-like object, opened in 'wb' mode
    :param url: Fully qualified URL to get the image from
    :param timeout: The timeout for the request
    :return:
    :raises WebcamError:
    """
    try:
        response = requests.get(url, allow_redirects=True, verify=False, timeout=timeout)
    except Exception as e:
        log.error(repr(e))
        raise WebcamError("Failed to request webcam URL {}".format(url))

    # The free space check before dumping to disk was previously commented in SP-RPI-SW, it is not implemented

    if response.status_code == 200:
        try:
            file_object.write(response.content)
        except Exception as e:
            log.error("Failed to save webcam image")
            log.exception(e)
            raise WebcamError("Failed to create file locally")
    else:
        raise WebcamError("Status code from download URL is not 200, was {} instead".format(response.status_code))


def start_livestream(queue, old_thread=None):
    if settings().get(["plugins", "SimplyPrint", "debug_logging"]):
        log.setLevel(logging.DEBUG)

    if old_thread is not None and old_thread.is_alive():
        queue.put("KILL")
        old_thread.join()

    new_queue = Queue()
    t = threading.Thread(target=livestream_loop, name="SimplyPrint webcam stream loop", args=(new_queue,))
    t.daemon = True
    t.start()
    return t


def livestream_loop(queue):
    fails = 0
    every = 1

    log.info("Starting streaming to server")
    log.info("Streaming to; " + WEBCAM_SNAPSHOT_URL)

    while True:
        if fails >= 10 or not queue.empty():
            # Failed to start, or queue message, stop and give up
            break

        start = time.time()
        try:
            request = post_image()
        except WebcamError:
            log.error("Error sending image for livestream")
            fails += 1
            # Sleep to prevent spam/overload
            time.sleep(0.8)
            continue

        if request is None:
            # Error, no response was returned, try again next time
            fails += 1
            time.sleep(0.8)
            continue

        try:
            response = request.json()
        except ValueError:
            log.error("Failed to parse response")
            fails += 1
            # Sleep to prevent spam/overload
            time.sleep(0.8)
            continue

        if "livestream" in response and response["livestream"] is not None:
            if not response["livestream"]["active"]:
                log.debug("Server stopped requesting for livestream")
                break

            every = response["livestream"]["every"]

        request_time = time.time() - start

        log.debug("webcam image took {}".format(request_time))

        if not queue.empty():
            # Message in the queue, need to stop
            break

        # Sleep until it is time to go again
        time.sleep(max((1 / every) - request_time, 0))


class WebcamError(Exception):
    """Standard webcam-related exception"""
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

import requests
try:
    # Python 3
    from urllib import parse as urlparse
except ImportError:
    import urlparse


class OctoPrintClient:
    """
    OctoPrint API Client wrapping the things SP uses
    Inspired by https://github.com/hroncok/octoclient
    """
    def __init__(self, url, api_key):
        self._logger = logging.getLogger(__name__)

        if not url:
            raise TypeError("Required argument 'url' not found or empty")
        if not api_key:
            raise TypeError("Required argument 'apikey' not found or empty")

        parsed = urlparse.urlparse(url)
        if parsed.scheme not in ["http", "https"]:
            raise TypeError("Provided URL is not http(s)")
        if not parsed.netloc:
            raise TypeError("Provided URL is empty")

        self.url = "{}://{}".format(parsed.scheme, parsed.netloc)
        self.session = requests.Session()
        self.session.headers.update({"X-Api-Key": api_key})

    def _check_response(self, response):
        """
        Make sure response is 20x
        :param response: requests.Response
        :return: response
        """

        if not (200 <= response.status_code < 210):
            error = response.text
            msg = "Response from {} was not OK: {} ({})".format(response.url, error, response.status_code)
            self._logger.error(msg)
            raise OctoPrintApiError(msg)

        return response

    def _get(self, path, params=None):
        """
        Http GET to OctoPrint at the path specified
        """
        url = urlparse.urljoin(self.url, path)
        response = self.session.get(url, params=params, timeout=5)
        self._check_response(response)
        return response.json()

    def _post(self, path, data=None, json=None, ret=True):
        url = urlparse.urljoin(self.url, path)
        response = self.session.post(url, data=data, json=json, timeout=5)
        self._check_response(response)

        if ret:
            return response.json()

    def version(self):
        return self._get("/api/version")

    def server(self):
        return self._get("/api/server")

    def settings(self, settings=None):
        if settings:
            return self._post("/api/settings", json=settings)
        else:
            return self._get("/api/settings")

    def plugin_plugin_manager(self):
        return self._get("/api/plugin/pluginmanager")

    def index(self):
        try:
            self._get("/")
        except ValueError:
            # Expected result, since it is not a JSON endpoint
            # Any other error implies that it is not available
            return True
        except Exception:
            return False


class OctoPrintApiError(Exception):
    """Basic error for a bad response from the API"""
    pass




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import socket
import sys
import threading
import io
import os

from octoprint.util.commandline import CommandlineCaller, CommandlineError

from .util import url_quote

GET_IP_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "get_ip.sh")


class SimplyPrintStartup:
    def __init__(self, simply_print):
        self.simply_print = simply_print
        self._logger = logging.getLogger("octoprint.plugins.SimplyPrint.comm.startup")
        self.startup_thread = None
        self.command_line = CommandlineCaller()

    def run_startup(self):
        if self.startup_thread is not None and self.startup_thread.is_alive():
            # Startup is in progress, lets leave it this way
            return

        thread = threading.Thread(target=self.startup)
        thread.daemon = True
        thread.start()
        self.startup_thread = thread

    def startup(self):
        # Filter out None values, switch to ''
        no_none = lambda x: x if x is not None else ""

        ip = no_none(self.get_ip())
        pi_model = no_none(self.get_pi_model())
        ssid = no_none(self.get_wifi())
        hostname = no_none(self.get_hostname())
        octoprint_version, octoprint_api_version = self.get_octoprint_version()
        python_version = no_none(self.get_python_version_str())

        public_port = self.get_public_port()
        if public_port and public_port != "" and public_port != "80":
            ip = "{}:{}".format(ip, public_port)

        url = "&startup=true" \
              "&device_ip={}" \
              "&pi_model={}" \
              "&wifi_ssid={}" \
              "&hostname={}" \
              "&octoprint_version={}" \
              "&octoprint_api_version={}" \
              "&python_version={}".format(
            url_quote(ip), url_quote(pi_model), url_quote(ssid), url_quote(hostname), url_quote(octoprint_version),
            url_quote(octoprint_api_version), url_quote(python_version)
        )

        request = self.simply_print.ping(url)

    @staticmethod
    def get_hostname():
        return socket.gethostname()

    @staticmethod
    def get_python_version_str():
        version_info = sys.version_info
        return "{version_info[0]}.{version_info[1]}.{version_info[2]}".format(**locals())

    @staticmethod
    def get_pi_model():
        try:
            with io.open("/proc/device-tree/model", "rt", encoding="utf-8") as file:
                return file.readline().strip(" \t\r\n\0")
        except:
            return

    def get_wifi(self):
        def iwgetid(command_path):
            try:
                returncode, stdout, stderr = self.command_line.checked_call([command_path, "-r"])
            except CommandlineError:
                raise

            return stdout[0].strip("\r\n")

        def iwlist(command_path):
            try:
                returncode, stdout, stderr = self.command_line.checked_call([command_path, "wlan0", "scan"])
            except CommandlineError:
                raise

            for line in stdout:
                line = line.lstrip()
                if line.startswith("ESSID"):
                    return line.split('"')[1]

        if os.path.exists("/usr/sbin/iwgetid"):
            command_path_iwgetid = "/usr/sbin/iwgetid"
        elif os.path.exists("/sbin/iwgetid"):
            command_path_iwgetid = "/sbin/iwgetid"
        else:
            self._logger.debug("iwgetid command not found")
            return None
        if os.path.exists("/usr/sbin/iwlist"):
            command_path_iwlist = "/usr/sbin/iwlist"
        elif os.path.exists("/sbin/iwlist"):
            command_path_iwlist = "/sbin/iwlist"
        else:
            self._logger.debug("iwlist command not found")
            return None

        try:
            self._logger.debug("iwgetid found at %s" % command_path_iwgetid)
            ssid = iwgetid(command_path_iwgetid)
            self._logger.debug("ssid: %s" % ssid)
        except CommandlineError:
            self._logger.warning("iwgetid failed")
            ssid = None

        if not ssid:
            try:
                self._logger.debug("iwlist found at %s" % command_path_iwlist)
                ssid = iwlist(command_path_iwlist)
                self._logger.debug("ssid: %s" % ssid)
            except CommandlineError:
                self._logger.warning("iwlist failed, can't get SSID")
                return None

        return ssid

    def get_ip(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            ip = s.getsockname()[0]
        except Exception:
            ip = ""

        if not ip or ip is None or ip == "127.0.1.1":
            try:
                returncode, stdout, stderr = self.command_line.checked_call(["bash", GET_IP_PATH])
            except CommandlineError:
                return None
            ip = stdout.strip("\r\n ").replace("\n", "")

        return ip

    @staticmethod
    def get_octoprint_version():
        """
        Get OctoPrint version and API version
        :return: (tuple) OctoPrint version, API version
        """
        from octoprint.server.api import VERSION
        from octoprint import __version__
        return __version__, VERSION

    def get_public_port(self):
        # noinspection PyProtectedMember
        return self.simply_print._settings.get(["public_port"])



# -*- coding: utf-8 -*-
#
# SimplyPrint
# Copyright (C) 2020-2022  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations
import requests
import pathlib
import logging
import tempfile
import time
import re
from threading import Event as ThreadEvent
from tornado.ioloop import IOLoop
from tornado.escape import url_escape, url_unescape

from octoprint.filemanager.util import DiskFileWrapper
from octoprint.filemanager.destinations import FileDestinations
from octoprint.events import eventManager, Events


from typing import (
    TYPE_CHECKING,
    Tuple,
    Awaitable,
    Dict,
    List,
    Any,
)
if TYPE_CHECKING:
    from .simplyprint import SimplyPrintWebsocket


def escape_query_string(qs: str) -> str:
    parts = qs.split("&")
    escaped: List[str] = []
    for p in parts:
        item = p.split("=", 1)
        key = url_escape(item[0])
        if len(item) == 2:
            escaped.append(f"{key}={url_escape(item[1])}")
        else:
            escaped.append(key)
    return "&".join(escaped)


class SimplyPrintFileHandler:
    def __init__(self, socket: SimplyPrintWebsocket) -> None:
        self.socket = socket
        self.file_manager = socket.file_manager
        self.printer = socket.printer
        self._logger = logging.getLogger("octoprint.plugins.simplyprint")
        self.pending_file: str = ""
        self.download_progress = -1
        self.start_event = ThreadEvent()
        self.analysis_event = ThreadEvent()
        self.start_event.set()
        self.analysis_event.set()

    def download_file(self, url: str, start: bool) -> Awaitable:
        if not hasattr(self, "_loop"):
            self._loop = IOLoop.current()
        return self._loop.run_in_executor(
            None, self._download_sp_file, url, start
        )

    def _escape_url(self, url: str) -> str:
        # escape the url
        match = re.match(r"(https?://[^/?#]+)([^?#]+)?(\?[^#]+)?(#.+)?", url)
        if match is not None:
            uri, path, qs, fragment = match.groups()
            if path is not None:
                uri += "/".join([url_escape(p, plus=False)
                                 for p in path.split("/")])
            if qs is not None:
                uri += "?" + escape_query_string(qs[1:])
            if fragment is not None:
                uri += "#" + url_escape(fragment[1:], plus=False)
            url = uri
        return url

    def _parse_content_disposition(self, data: str) -> str:
        fnr = r"filename[^;\n=]*=(['\"])?(utf-8\'\')?([^\n;]*)(?(1)\1|)"
        matches: List[Tuple[str, str, str]] = re.findall(fnr, data)
        is_utf8 = False
        filename: str = ""
        for (_, encoding, fname) in matches:
            if encoding.startswith("utf-8"):
                # Prefer the utf8 filename if included
                filename = url_unescape(
                    fname, encoding="utf-8", plus=False)
                is_utf8 = True
                break
            filename = fname
        self._logger.debug(
            "Content-Disposition header received: filename = "
            f"{filename}, utf8: {is_utf8}"
        )
        return filename

    def _download_sp_file(self, url: str, start: bool):
        tmp_fname = f"sp-{time.monotonic_ns()}.gcode"
        tmp_path = pathlib.Path(tempfile.gettempdir()).joinpath(tmp_fname)
        size: int = 0
        downloaded: int = 0
        last_pct: int = 0
        url_parts = url.rsplit("/", 1)
        filename: str = tmp_fname
        if len(url_parts) == 2 and "." in url_parts[-1]:
            # The last fragment of the path looks like a filename,
            # initialize to that.
            filename = url_parts[-1]
        url = self._escape_url(url)
        self.download_progress = -1
        try:
            kwargs = dict(allow_redirects=True, stream=True, timeout=3600.)
            with requests.get(url, **kwargs) as resp:
                resp.raise_for_status()
                self._update_progress(0)
                size = int(resp.headers.get("content-length", 0))
                if "content-disposition" in resp.headers:
                    cd = resp.headers["content-disposition"]
                    fname = self._parse_content_disposition(cd)
                    if fname:
                        filename = fname
                with tmp_path.open("wb") as f:
                    for chunk in resp.iter_content(8192):
                        downloaded += len(chunk)
                        f.write(chunk)
                        if size:
                            pct = int(downloaded / size * 100 + .4)
                            if pct != last_pct:
                                last_pct = pct
                                self._update_progress(pct)
        except Exception as e:
            self._logger.exception("Error downloading print")
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "error", "message": "Network Error", "exception": e}
            )
            return
        local = FileDestinations.LOCAL
        if not self.file_manager.folder_exists(local, "SimplyPrint"):
            self.file_manager.add_folder(local, "SimplyPrint")
        count: int = 0
        while count < 100:
            try:
                cpath, cname = self.file_manager.canonicalize(
                    local, f"SimplyPrint/{filename}"
                )
                folder = self.file_manager.sanitize_path(local, cpath)
                name = self.file_manager.sanitize_name(local, cname)
                dest_path = self.file_manager.join_path(local, folder, name)
                storage_path = self.file_manager.path_in_storage(local, dest_path)
                if self.printer.can_modify_file(storage_path, False):
                    fobj = DiskFileWrapper(tmp_fname, str(tmp_path))
                    added_file = self.file_manager.add_file(
                        local, storage_path, fobj, allow_overwrite=True,
                        display=cname
                    )
                    break
                count += 1
                # This file name won't work, try another one
                fparts = filename.split(".", 1)
                ext = "gcode" if len(fparts) < 2 else fparts[-1]
                filename = f"{fparts[0]}_copy{count}.{ext}"
            except Exception as e:
                self._logger.exception("Error locating file destination")
                self._loop.add_callback(
                    self.socket.send_sp, "file_progress",
                    {"state": "error", "message": "Error processing download", "exception": e}
                )
                return
        else:
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "error", "message": "Error processing download"}
            )
            return
        self.analysis_event.clear()
        try:
            self.printer.select_file(storage_path, False, False)
        except Exception:
            pass
        selected = self.printer.is_current_file(storage_path, False)
        # Fire the upload event
        eventManager().fire(
            Events.UPLOAD,
            {
                "name": name,
                "path": added_file,
                "target": local,
                "select": selected,
                "print": False
            }
        )
        if not selected:
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "error", "message": "Error selecting file"}
            )
            self.analysis_event.set()
            return

        self.analysis_event.wait(5.)
        self.analysis_event.set()
        self.pending_file = storage_path
        if start:
            self.start_print()
        else:
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "ready"}
            )

    def start_print(self) -> bool:
        if not hasattr(self, "_loop"):
            # File has never been uploaded or selected
            return False
        if not self.pending_file:
            return False
        pending = self.pending_file
        self.pending_file = ""
        if not self.printer.is_current_file(pending, False):
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "error", "message": "File not selected"}
            )
            return False
        self.start_event.clear()
        self.printer.start_print()
        printing = self.start_event.wait(10.)
        self.start_event.set()
        if not printing:
            self._loop.add_callback(
                self.socket.send_sp, "file_progress",
                {"state": "error", "message": "Error starting print"}
            )
            return False
        return True

    def start_pending(self) -> bool:
        return not self.start_event.is_set()

    def notify_started(self) -> None:
        self.start_event.set()

    def _update_progress(self, percent: int) -> None:
        if percent == self.download_progress:
            return
        self.download_progress = percent
        self._loop.add_callback(
            self.socket.send_sp, "file_progress",
            {"state": "downloading", "percent": percent}
        )

    def check_analysis(self, payload: Dict[str, Any]) -> None:
        if self.analysis_event.is_set():
            return
        if self.printer.is_current_file(payload["path"], False):
            self.analysis_event.set()


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This script should be run in the background on the Raspberry Pi, to check that OctoPrint is alive and if it is not alive
then it will let SP know that it has died, and await instruction on how to proceed.
"""
import logging
import threading
import time

import requests

from octoprint.settings import settings
from octoprint.util import ResettableTimer
from octoprint.util.commandline import CommandlineCaller

from octoprint_simplyprint.comm.constants import UPDATE_URL, API_VERSION
from octoprint_simplyprint.local.util import OctoPrintClient, OctoPrintApiError


def run_background_check():
    simply_background = SimplyPrintBackground()
    simply_background.mainloop()


class SimplyPrintBackground:
    def __init__(self):
        self._logger = logging.getLogger("octoprint.plugins.SimplyPrint.background")
        self._logger.setLevel(logging.DEBUG)

        try:
            self._octoprint_settings = settings(init=True)
            # We need init as this runs in a separate process
            # NOTE: This should not be used to write to the settings file, since it could cause a conflict
            # with OctoPrint's settings, and lose some settings there.
        except ValueError:
            self._logger.error("This script shouldn't be run in the same process as OctoPrint")
            self._logger.error("So don't do that :) ")
            return

        self.octoprint = None

        self.was_octoprint_up = False
        self.failed_checks = 0

        self.main_thread = None
        self.run = True

        self.safe_mode_checks = 0

    def mainloop(self):
        port = self._octoprint_settings.get(["public_port"])
        if port is not None and port != "" and port != "80":
            ip = "http://127.0.0.1:{}".format(port)
        else:
            ip = "http://127.0.0.1"

        self.octoprint = OctoPrintClient(ip, self._octoprint_settings.get(["api", "key"]))

        while self.run:
            try:
                start = time.time()

                check_result = self.check_octoprint()
                if not check_result and self.was_octoprint_up:  # Only restart if OctoPrint was previously up
                    self._logger.warning("OctoPrint is not OK...")
                    self.ping_simplyprint("&octoprint_status=Shutdown")
                    self.failed_checks += 1
                    if self.failed_checks >= 2:
                        # Only restart after 2 consecutive failed checks
                        self._logger.warning("Trying to restart it now")
                        self.restart_octoprint()
                        self.failed_checks = 0

                elif not check_result and not self.was_octoprint_up:
                    self._logger.warning("OctoPrint hasn't been seen yet, skipping")

                else:
                    self._logger.debug("OctoPrint seems OK")
                    self.was_octoprint_up = True
                    self.failed_checks = 0

                    safe_mode = self.check_safemode()
                    if not safe_mode:
                        self._logger.warning("OctoPrint is in safe mode")
                        if self.safe_mode_checks == 0 or self.safe_mode_checks > 10:
                            # Restart immediately, or after more than 10 mins
                            self.restart_octoprint()
                            self.safe_mode_checks = 0

                    else:
                        self._logger.debug("OctoPrint is not in safe mode")

                    self.safe_mode_checks += 1

                total_time = time.time() - start
                self._logger.debug("OctoPrint health check took {}".format(total_time))
                if self.run:
                    time.sleep(60 - total_time)
            except Exception as e:
                self._logger.exception(e)
                time.sleep(60)

    def check_octoprint(self):
        """
        Checks OctoPrint is alive
        """
        def check_api():
            try:
                version = self.octoprint.version()
            except OctoPrintApiError:
                return False
            except Exception:
                # Some other random error, possibly connection refused if proxy dies
                return False

            if "octoprint" in version["text"].lower():
                return True
            else:
                return False

        def check_index():
            # This checks for the intermediary server too, if the API is offline
            # Means it is probably booting
            try:
                return self.octoprint.index()
            except Exception:
                return False

        if not check_api():
            return check_index()
        else:
            return True

    def check_safemode(self):
        def check_server():
            try:
                server = self.octoprint.server()
            except OctoPrintApiError:
                # OctoPrint < 1.5.0, no /api/server
                return False

            return server["safemode"] is None

        def check_pgmr():
            try:
                pgmr = self.octoprint.plugin_plugin_manager()
            except OctoPrintApiError:
                # Now it's possible its dead, more likely that user disabled plugin manager :/
                # Return True since its not definitive
                return True

            for plugin in pgmr["plugins"]:
                if plugin["safe_mode_victim"]:
                    return False

            return True

        if not check_server():
            return check_pgmr()
        else:
            return True

    def restart_octoprint(self):
        command = self._octoprint_settings.get(["server", "commands", "serverRestartCommand"])
        if not command:
            self._logger.warning("No command configured, can't restart")
            return

        caller = CommandlineCaller()
        try:
            code, stdout, stderr = caller.call(command, **{"shell": True}) # Use shell=True, as we have to trust user input
        except Exception as e:
            self._logger.error("Error calling command to restart server {}".format(command))
            self._logger.exception(e)
            return

        if code != 0:
            self._logger.error("Non zero return code running '{}' to restart server: {}".format(command, code))
            self._logger.exception("STDOUT: {}".format(stdout))
            self._logger.exception("STDERR: {}".format(stderr))

    def ping_simplyprint(self, parameters):
        rpi_id = self._octoprint_settings.get(["plugins", "SimplyPrint", "rpi_id"])
        if not rpi_id:
            # Not set up properly, nothing we can do - let the plugin handle getting a new ID
            return

        url = UPDATE_URL + "?id=" + rpi_id + "&api_version=" + API_VERSION
        url = url.replace(" ", "%20")

        try:
            response = requests.get(url, timeout=5)
        except requests.exceptions.RequestException as e:
            self._logger.error("Error sending get request to SimplyPrint")
            self._logger.exception(e)
            raise

        return response


if __name__ == '__main__':
    run_background_check()



# -*- coding: utf-8 -*-
#
# SimplyPrint
# Copyright (C) 2020-2022  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations
import asyncio
import logging
import requests
import base64
from tornado.ioloop import IOLoop

from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
)

if TYPE_CHECKING:
    from octoprint.plugin import PluginSettings

class WebcamStream:
    def __init__(
        self, settings: PluginSettings, image_callback: Callable[[str], None]
    ) -> None:
        self._settings = settings
        self._logger = logging.getLogger("octoprint.plugins.simplyprint")
        self.url: str = settings.global_get(["webcam", "snapshot"])
        self.running = False
        self.interval: float = 1.
        self.on_image_received = image_callback
        self.stream_task: Optional[asyncio.Task] = None
        self._connection_test_passed = False

    @property
    def webcam_connected(self):
        return self._connection_test_passed

    async def test_connection(self) -> None:
        if not hasattr(self, "_loop"):
            self._loop = IOLoop.current()
        img = await self._loop.run_in_executor(None, self.extract_image)
        self._connection_test_passed = img is not None

    def extract_image(self) -> Optional[str]:
        headers = {"Accept": "image/jpeg"}
        try:
            resp = requests.get(
                self.url, headers=headers, verify=False, timeout=4
            )
            resp.raise_for_status()
        except Exception:
            return None
        return self._encode_image(resp.content)

    def _encode_image(self, image: bytes) -> str:
        return base64.b64encode(image).decode()

    async def _stream(self) -> None:
        while self.running:
            try:
                ret = await self._loop.run_in_executor(
                    None, self.extract_image
                )
                if ret is not None:
                    self.on_image_received(ret)
                await asyncio.sleep(self.interval)
            except asyncio.CancelledError:
                break

    def start(self, interval: float) -> None:
        if not self.url.startswith("http"):
            self._logger.info(
                f"Invalid webcam url, aborting stream: {self.url}"
            )
            return
        if self.running:
            return
        if not hasattr(self, "_loop"):
            self._loop = IOLoop.current()
        self.interval = interval
        self.running = True
        self.stream_task = self._loop.add_callback(self._stream)

    def stop(self) -> None:
        if not self.running:
            return
        self.running = False
        if self.stream_task is not None:
            self.stream_task.cancel()
            self.stream_task = None


# -*- coding: utf-8 -*-
#
# SimplyPrint
# Copyright (C) 2020-2022  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations
import asyncio
import json
import pathlib
import logging
import functools
import threading
import re
import tornado.websocket
from octoprint.util import server_reachable
from tornado.ioloop import IOLoop

from .constants import WS_TEST_ENDPOINT, WS_PROD_ENDPOINT
from .system import SystemQuery, SystemManager
from .webcam import WebcamStream
from .file_handler import SimplyPrintFileHandler
from ..comm.monitor import Monitor
import octoprint.server
import octoprint.util
from octoprint.plugin import PluginSettings
from octoprint.printer import PrinterInterface
from octoprint.filemanager import FileManager, FileDestinations
from octoprint.events import Events, EventManager
import requests
import datetime

# XXX: The below imports are for inital dev and
# debugging.  They are used to create a logger for
# messages sent to and received from the simplyprint
# backend
import logging.handlers
from queue import SimpleQueue

from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Awaitable,
    Dict,
    List,
    Tuple,
    Union,
    Any,
    cast,
)
if TYPE_CHECKING:
    from tornado.websocket import WebSocketClientConnection
    from ..import SimplyPrint
    TimerCallback = Callable[[float], Union[float, Awaitable[float]]]


UPDATE_CHECK_TIME = 24. * 60. * 60
# TODO: Increase this time to something greater, perhaps 30 minutes
CONNECTION_ERROR_LOG_TIME = 60. * 60
VALID_STATES = [
    "offline",  "operational", "printing", "cancelling",
    "pausing", "paused", "resuming", "error"
]
PRE_SETUP_EVENTS = [
    "connection", "state_change", "shutdown", "machine_data", "keepalive",
    "firmware", "installed_plugins", "firmware_warning"
]

class SimplyPrintWebsocket:
    def __init__(self, plugin: SimplyPrint) -> None:
        self.plugin = plugin
        self._logger = logging.getLogger("octoprint.plugins.simplyprint")
        self.settings = cast(PluginSettings, plugin._settings)
        self.printer = cast(PrinterInterface, plugin._printer)
        self.file_manager = cast(FileManager, plugin._file_manager)
        self.sys_manager = SystemManager(self)
        self._event_bus = cast(EventManager, plugin._event_bus)
        if self.settings.get_boolean(["debug_logging"]):
            self._logger.setLevel(logging.DEBUG)
        self.test = False
        if self.settings.get(["endpoint"]) == "test":
            self.test = True
        self.connected = False
        self.is_set_up = self.settings.get(["is_set_up"])
        self._set_ws_url()

        self.simplyprint_thread = threading.Thread(
            target=self._run_simplyprint_thread
        )
        self.simplyprint_thread.daemon = True
        self.is_closing = False
        self.is_connected = False
        self._user_input_req = False
        self.ws: Optional[WebSocketClientConnection] = None
        self.cache = ReportCache()
        self.current_layer: int = -1
        self.last_received_temps: Dict[str, float] = {}
        self.last_err_log_time: float = 0.
        self.download_progress: int = -1
        self.intervals: Dict[str, float] = {
            "job": 5.,
            "temps": 5.,
            "temps_target": 2.5,
            "cpu": 30.,
            "reconnect": 30.,
            "ai": 60.,
            "ready_message": 60.,
            "ping": 20.
        }
        self.temp_timer = FlexTimer(self._handle_temperature_update)
        self.job_info_timer = FlexTimer(self._handle_job_info_update)
        self.cpu_timer = FlexTimer(self._handle_cpu_update)
        self.printer_reconnect_timer = FlexTimer(self._handle_printer_reconnect)
        self.update_timer = FlexTimer(self._handle_update_check)
        self.monitor = Monitor(
            logging.getLogger("octoprint.plugins.simplyprint.monitor")
        )
        self.ai_timer = FlexTimer(self._handle_ai_snapshot)
        self.failed_ai_attempts = 0
        self.scores = []
        self.ping_timer = FlexTimer(self._handle_ping)
        self._last_ping_sent: float = 0.
        self.reset_printer_display_timer = FlexTimer(self._reset_printer_display)
        self.webcam_stream = WebcamStream(self.settings, self._send_image)
        self._image_delivered = True
        self.amb_detect = AmbientDetect(
            self.cache, self._on_ambient_changed,
            self.settings.get_int(["ambient_temp"])
        )
        self.file_handler = SimplyPrintFileHandler(self)
        self.heaters: Dict[str, str] = {}
        self.missed_job_events: List[Dict[str, Any]] = []
        self.connection_task: Optional[asyncio.Task] = None
        self.reconnect_delay: float = 1.
        self.reconnect_token: Optional[str] = None
        self._last_ping_received: float = 0.
        self.gcode_terminal_enabled: bool = False
        self.cached_events: List[Tuple[Callable, Tuple[Any, ...]]] = []

        # XXX: The call below is for dev, remove before release
        self._setup_simplyprint_logging()


    ######################################################################
    #
    # Plugin Methods / Properties
    #
    # These methods are defined in the original "SimplyPrintComm" class
    # and called by the SimplyPrint plugin.  When it is no longer required
    # to support both they can be refactored.
    #
    ######################################################################

    def get_helpers(self, plugin_name: str) -> Optional[Dict[str, Callable]]:
        pm = self.plugin._plugin_manager
        return pm.get_helpers(plugin_name)  # type: ignore

    def on_startup(self) -> None:
        if self.simplyprint_thread.is_alive():
            return
        self.simplyprint_thread.start()

    def _run_simplyprint_thread(self) -> None:
        self._aioloop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._aioloop)
        self._aioloop.run_until_complete(self._initialize())

    def on_layer_change(self, layer: int) -> None:
        self.current_layer = layer

    def on_pause_at_command(self, message: str) -> None:
        # TODO: Not sure what to submit here
        pass

    def on_gcode_received(self, line: str) -> None:
        if not hasattr(self, "_loop"):
            return
        if self.gcode_terminal_enabled:
            self._loop.add_callback(
                self.send_sp, "term_update", {"response": [line]}
            )

    def on_gcode_sent(self, cmd: str) -> None:
        if not hasattr(self, "_loop"):
            return
        if self.gcode_terminal_enabled:
            self._loop.add_callback(
                self.send_sp, "term_update", {"command": [cmd]}
            )

    def _queue_event(self, func: Callable, *args) -> None:
        self.cached_events.append((func, args))

    def on_event(self, event: str, payload: Dict[str, Any]) -> None:
        if not hasattr(self, "_loop"):
            add_callback = self._queue_event
        else:
            add_callback = self._loop.add_callback

        if event == Events.PRINTER_STATE_CHANGED:
            state = payload["state_id"].lower()
            if state in ["operational", "printing", "pausing", "resuming"]:
                add_callback(self._on_state_event, state)
        elif event in [Events.CONNECTING, Events.DISCONNECTING]:
            add_callback(self._send_connection_state, event.lower())
        elif event == Events.CONNECTED:
            add_callback(self._on_printer_connected)
        elif event == Events.DISCONNECTED:
            add_callback(self._on_printer_disconnected)
        elif event == Events.ERROR:
            msg = payload.get("error", "")
            add_callback(self._on_printer_error, msg)
        elif event == Events.PRINT_STARTED:
            add_callback(self._on_print_start, payload)
        elif event == Events.PRINT_FAILED:
            if payload.get("reason", "") == "error":
                add_callback(self._on_print_done, "failed")
        elif event == Events.PRINT_DONE:
            add_callback(self._on_print_done, "finished")
        elif event == Events.PRINT_CANCELLING:
            msg = payload.get("firmwareError", "")
            add_callback(self._on_cancelling_event, msg)
        elif event == Events.PRINT_PAUSED:
            add_callback(self._on_print_paused)
        elif event == Events.PRINT_RESUMED:
            add_callback(self._on_print_resumed)
        elif event == Events.TOOL_CHANGE:
            if "new" not in payload:
                return
            add_callback(self._send_active_extruder, payload["new"])
        elif event == Events.FIRMWARE_DATA:
            add_callback(self._send_firmware_data, payload)
        elif event == Events.METADATA_ANALYSIS_FINISHED:
            add_callback(self._on_metadata_update, payload)
        elif event == "plugin_firmware_check_warning":
            add_callback(self._on_firmware_warning, payload)
        elif event == "plugin_printer_safety_check_warning":
            add_callback(self._on_firmware_warning, payload)
        elif event == "plugin_pi_support_throttle_state":
            add_callback(self._on_cpu_throttled, payload)
        elif event == "plugin_bedlevelvisualizer_mesh_data_collected":
            add_callback(self._send_mesh_data, payload["mesh"])
        elif event == Events.FILE_SELECTED:
            # TODO: not sure we need this one
            pass
        elif event == Events.FILE_REMOVED:
            # TODO: not sure we need this one
            pass
        elif event == "plugin_simplyfilamentsensor_filament_loaded":
            add_callback(self.send_sp, "filament_sensor", {"state": "loaded"})
        elif event == "plugin_simplyfilamentsensor_filament_runout":
            add_callback(self.send_sp, "filament_sensor", {"state": "runout"})
        elif event == "plugin_simplyfilamentsensor_filament_no_filament_print_on_print_start":
            add_callback(self.send_sp, "filament_sensor", {"state": "print_stopped"})
        elif event == "plugin_psucontrol_psu_state_changed":
            is_on = payload["isPSUOn"]
            add_callback(self.send_sp, "power_controller", {"on": is_on})
        elif event == "plugin_simplypowercontroller_power_on":
            add_callback(self.send_sp, "power_controller", {"on": True})
        elif event == "plugin_simplypowercontroller_power_off":
            add_callback(self.send_sp, "power_controller", {"on": False})
        if event in [
            "plugin_pluginmanager_install_plugin",
            "plugin_pluginmanager_uninstall_plugin",
            "plugin_pluginmanager_enable_plugin",
            "plugin_pluginmanager_disabled_plugin",
        ]:
            if event == "plugin_pluginmanager_uninstall_plugin":
                if payload.get("id", "") == "SimplyPrint":
                    self._logger.info("The SimplyPrint plugin was uninstalled")
            add_callback(self._send_installed_plugins)

    @property
    def user_input_required(self) -> bool:
        return self._user_input_req

    @user_input_required.setter
    def user_input_required(self, state: bool) -> None:
        if self._user_input_req != state:
            self._user_input_req = state
            if not self.printer.is_printing():
                return
            new_state = "paused" if state else "printing"
            self._loop.add_callback(self._update_state, new_state)

    def _monotonic(self) -> float:
        return self._aioloop.time()

    async def _initialize(self) -> None:
        self._loop = IOLoop.current()
        for (func, args) in self.cached_events:
            func(*args)
        self.cached_events = []
        self.update_timer.start(delay=60.)
        self.cpu_timer.start()
        self.ping_timer.start()
        if self.printer.is_operational():
            self._on_printer_connected()
        else:
            self._update_state_from_octo()
        self.cache.machine_data = await self._get_machine_data()
        await self.webcam_stream.test_connection()
        await self._connect()

    async def _get_machine_data(self) -> Dict[str, Any]:
        sys_query = SystemQuery(self.settings)
        data = await self._loop.run_in_executor(None, sys_query.get_system_info)
        data["ui"] = "OctoPrint"
        data["api"] = "OctoPrint"
        data["sp_version"] = self.plugin._plugin_version
        return data

    async def _connect(self) -> None:
        log_connect = True
        failed_attempts = 0
        while not self.is_closing:
            url = self.connect_url
            if self.reconnect_token is not None:
                url = f"{self.connect_url}/{self.reconnect_token}"
            if log_connect:
                self._logger.info(f"Connecting To SimplyPrint: {url}")
                log_connect = False
            url_start = self.connect_url[6:]
            url_start_match = re.match(r"wss://([^/]+)", self.connect_url)
            if url_start_match is not None:
                url_start = url_start_match.group(1)
            try:
                reachable = await self._loop.run_in_executor(
                    None, server_reachable, url_start, 80
                )
                if not reachable:
                    raise Exception("SimplyPrint not Reachable")
                self.ws = await tornado.websocket.websocket_connect(
                    url, connect_timeout=15.
                )
                setattr(self.ws, "on_ping", self._on_ws_ping)
                cur_time = self._monotonic()
                self._last_ping_received = cur_time
            except asyncio.CancelledError:
                raise
            except Exception:
                curtime = self._monotonic()
                timediff = curtime - self.last_err_log_time
                if timediff > CONNECTION_ERROR_LOG_TIME:
                    self.last_err_log_time = curtime
                    self._logger.info(
                        f"Failed to connect to SimplyPrint")
                failed_attempts += 1
                if not failed_attempts % 10:
                    self.set_display_message("Can't reach SP", True)
                    self.reset_printer_display_timer.start(delay=120)
            else:
                if failed_attempts:
                    self.set_display_message("Back online!", True)
                failed_attempts = 0
                self._logger.info("Connected to SimplyPrint Cloud")
                self._start_printer_reconnect()
                await self._read_messages()
                log_connect = True
            if not self.is_closing:
                await asyncio.sleep(self.reconnect_delay)

    async def _read_messages(self) -> None:
        message: Union[str, bytes, None]
        while self.ws is not None:
            message = await self.ws.read_message()
            if isinstance(message, str):
                self._process_message(message)
            elif message is None:
                self.webcam_stream.stop()
                cur_time = self._monotonic()
                ping_time: float = cur_time - self._last_ping_received
                reason = code = None
                if self.ws is not None:
                    reason = self.ws.close_reason
                    code = self.ws.close_code
                msg = (
                    f"SimplyPrint Disconnected - Code: {code}, Reason: {reason}, "
                    f"Server Ping Time Elapsed: {ping_time}"
                )
                self._logger.info(msg)
                self.connected = False
                self.ws = None
                break

    def _on_ws_ping(self, data: bytes = b"") -> None:
        self._last_ping_received = self._monotonic()

    def _process_message(self, msg: str) -> None:
        self._sock_logger.info(f"received: {msg}")
        try:
            packet: Dict[str, Any] = json.loads(msg)
        except json.JSONDecodeError:
            self._logger.debug(f"Invalid message, not JSON: {msg}")
            return
        event: str = packet.get("type", "")
        data: Optional[Dict[str, Any]] = packet.get("data")
        if event == "connected":
            self._logger.info("SimplyPrint Reports Connection Success")
            self.connected = True
            self.reconnect_token = None
            if data is not None:
                if data.get("in_setup", 0) == 1:
                    self.is_set_up = False
                    self.settings.set_boolean(["is_set_up"], False)
                    self.settings.set(["printer_id"], "")
                    self.settings.set(["printer_name"], "")
                    if "short_id" in data:
                        self.settings.set(["temp_short_setup_id"], data["short_id"])
                    self.settings.save(trigger_event=True)
                interval = data.get("interval")
                self._set_intervals(interval)
                self.reconnect_token = data.get("reconnect_token")
                name = data.get("name")
                if name is not None:
                    self._save_item("printer_name", name)
            self.reconnect_delay = 1.
            self._push_initial_state()
        elif event == "error":
            self._logger.info(f"SimplyPrint Connection Error: {data}")
            self.reconnect_delay = 30.
            self.reconnect_token = None
        elif event == "new_token":
            if data is None:
                self._logger.debug("Invalid message, no data")
                return
            if data.get("no_exist", False) is True and self.is_set_up:
                self.is_set_up = False
                self.settings.set_boolean(["is_set_up"], False)
                self.settings.save()
            token = data.get("token")
            if not isinstance(token, str):
                self._logger.info(f"Invalid token received: {token}")
                return
            self._logger.info(f"SimplyPrint Token Received")
            self._save_item("printer_token", token)
            short_id = data.get("short_id")
            if not isinstance(short_id, str):
                self._logger.debug(f"Invalid short_id received: {short_id}")
            else:
                self.settings.set(["temp_short_setup_id"], data["short_id"])
                self.settings.save(trigger_event=True)
            self._set_ws_url()
        elif event == "complete_setup":
            if data is None:
                self._logger.debug("Invalid message, no data")
                return
            printer_id = data.get("printer_id")
            self.settings.set(["printer_id"], str(printer_id))
            self.is_set_up = True
            self.settings.set_boolean(["is_set_up"], True)
            self.settings.set(["temp_short_setup_id"], "")
            self.settings.save(trigger_event=True)
            self.set_display_message("Set up!")
            self._set_ws_url()
        elif event == "demand":
            if data is None:
                self._logger.debug(f"Invalid message, no data")
                return
            demand = data.pop("demand", "unknown")
            self._process_demand(demand, data)
        elif event == "interval_change":
            self._set_intervals(data)
        elif event == "pong":
            cur_time = self._monotonic()
            time_diff: float = cur_time - self._last_ping_sent
            self.send_sp("latency", {"ms": int(time_diff*1000)})
        elif event == "stream_received":
            self._image_delivered = True
        elif event == "printer_settings":
            if data is not None:
                self._sync_settings_from_simplyprint(data)
        else:
            # TODO: It would be good for the backend to send an
            # event indicating that it is ready to recieve printer
            # status.
            self._logger.debug(f"Unknown event: {msg}")

    def _set_intervals(self, data):
        if isinstance(data, dict):
            ai_timer_retart = False
            for key, val in data.items():
                if key == "ai" and val / 1000. < self.intervals.get("ai"):
                    ai_timer_retart = True
                self.intervals[key] = val / 1000.
            self._logger.debug(f"Intervals Updated: {self.intervals}")
            if ai_timer_retart and self.printer.is_printing():
                if not hasattr(self, "ai_timer_not_before"):
                    self.ai_timer_not_before = datetime.datetime.now() + datetime.timedelta(seconds=self.intervals.get("ai"))
                td = 0 if datetime.datetime.now() > self.ai_timer_not_before else 120. - (self.ai_timer_not_before - datetime.datetime.now()).total_seconds()
                self.ai_timer.stop()
                self.ai_timer.start(delay=td)
            self._start_printer_reconnect()

    def _process_demand(self, demand: str, args: Dict[str, Any]) -> None:
        if demand == "pause":
            self.set_display_message("Pausing...", True)
            self._update_state("pausing")
            self._loop.run_in_executor(None, self.printer.pause_print)
        elif demand == "resume":
            self._update_state("resuming")
            self.set_display_message("Resuming...", True)
            self._loop.run_in_executor(None, self.printer.resume_print)
        elif demand == "cancel":
            self._update_state("cancelling")
            self.set_display_message("Cancelling...", True)
            self._loop.run_in_executor(None, self.printer.cancel_print)
        elif demand == "terminal":
            if "enabled" in args:
                self.gcode_terminal_enabled = args["enabled"]
        elif demand == "gcode":
            script_list = args.get("list", [])
            if script_list:
                self._loop.run_in_executor(
                    None, self.printer.commands, script_list
                )
        elif demand == "test_webcam":
            self._loop.add_callback(self._test_webcam)
        # elif demand == "stream_on":
        #     self._image_delivered = True
        #     interval: float = args.get("interval", 1000) / 1000
        #     self.webcam_stream.start(interval)
        # elif demand == "stream_off":
        #     self._image_delivered = True
        #     self.webcam_stream.stop()
        elif demand == "webcam_snapshot":
            if self.webcam_stream.webcam_connected:
                self._loop.add_callback(self._post_snapshot, **args)
        elif demand == "file":
            self.set_display_message("Preparing...", True)
            url: Optional[str] = args.get("url")
            if not isinstance(url, str):
                self._logger.debug(f"Invalid url in message")
                return
            if self.file_manager.folder_exists(FileDestinations.LOCAL, "SimplyPrint"):
                files = self.file_manager.list_files(FileDestinations.LOCAL, "SimplyPrint")
                for file, data in files["local"].items():
                    self.file_manager.remove_file(FileDestinations.LOCAL, data["path"])
            start = bool(args.get("auto_start", 0))
            self.file_handler.download_file(url, start)
        elif demand == "start_print":
            def _on_start_finished(fut: asyncio.Future):
                if not fut.result():
                    self._logger.debug("Failed to start print")
            fut: asyncio.Future = self._loop.run_in_executor(  # type: ignore
                None, self.file_handler.start_print
            )
            fut.add_done_callback(_on_start_finished)
        elif demand == "connect_printer":
            if self.printer.is_closed_or_error():
                self._logger.info(
                    "Connecting to printer at request from SimplyPrint"
                )
                self._loop.run_in_executor(None, self.printer.connect)
        elif demand == "disconnect_printer":
            self._logger.info(
                "Disconnecting printer at request from SimplyPrint"
            )
            self._loop.run_in_executor(None, self.printer.disconnect)
        elif demand == "system_restart":
            self.set_display_message("Rebooting...", True)
            self._loop.run_in_executor(
                None, self.sys_manager.reboot_machine
            )
        elif demand == "system_shutdown":
            self.set_display_message("Shutting Down...", True)
            self._loop.run_in_executor(
                None, self.sys_manager.shutdown_machine
            )
        elif demand == "api_restart":
            self.set_display_message("Restarting OctoPrint", True)
            self._loop.run_in_executor(
                None, self.sys_manager.restart_octoprint
            )
        elif demand == "api_shutdown":
            self.set_display_message("Stopping OctoPrint...", True)
            self._loop.run_in_executor(
                None, self.sys_manager.stop_octoprint
            )
        elif demand == "update":
            self.set_display_message("Updating...", True)
            self._loop.run_in_executor(
                None, self.sys_manager.update_simplyprint
            )
        elif demand == "plugin_install":
            def _on_install_finished(install_success: asyncio.Future):
                if install_success.result():
                    self._logger.info("Restarting OctoPrint after plugin install(s).")
                    self._loop.run_in_executor(
                        None, self.sys_manager.restart_octoprint
                    )
                else:
                    self._logger.debug("Failed to install plugin")
            if args.get("plugins", False):
                install_success: asyncio.Future = self._loop.run_in_executor(  # type: ignore
                    None, self.sys_manager.install_plugin, args["plugins"]
                )
                install_success.add_done_callback(_on_install_finished)
        elif demand == "plugin_uninstall":
            def _on_uninstall_finished(uninstall_success: asyncio.Future):
                if uninstall_success.result():
                    self._logger.info("Restarting OctoPrint after plugin uninstall.")
                    self._loop.run_in_executor(
                        None, self.sys_manager.restart_octoprint
                    )
                else:
                    self._logger.debug("Failed to uninstall plugin")
            uninstall_success: asyncio.Future = self._loop.run_in_executor(  # type: ignore
                None, self.sys_manager.uninstall_plugin, args
            )
            uninstall_success.add_done_callback(_on_uninstall_finished)
        elif demand == "webcam_settings_updated":
            cam_settings = args.get("webcam_settings")
            if cam_settings is not None:
                self._sync_webcam_settings(cam_settings)
        elif demand == "set_printer_profile":
            profile = args.get("printer_profile")
            if profile is not None:
                self._loop.run_in_executor(
                    None, self._save_printer_profile, profile
                )
        elif demand == "get_gcode_script_backups":
            force = args.get("force", False)
            self._send_gcode_scripts(force)
        elif demand == "has_gcode_changes":
            scripts = args.get("scripts")
            if scripts is not None:
                self._save_gcode_scripts(scripts)
        elif demand in ["psu_on", "psu_keepalive"]:
            self._loop.run_in_executor(
                None, self.sys_manager.power_on_printer
            )
        elif demand == "psu_off":
            self._loop.run_in_executor(
                None, self.sys_manager.power_off_printer
            )
        elif demand == "disable_websocket":
            self._save_item("websocket_ready", False)
            self._loop.run_in_executor(
                None, self.sys_manager.restart_octoprint
            )
        else:
            self._logger.debug(f"Unknown demand: {demand}")

    def _sync_settings_from_simplyprint(
        self, sp_settings: Dict[str, Any]
    ) -> None:
        self._logger.info(
            "Syncing Settings at request from SimplyPrint"
        )
        if "display" in sp_settings:
            display = sp_settings["display"]
            if "enabled" in display:
                self.settings.set_boolean(
                    ["display_enabled"], display["enabled"] == 1
                )
            if "branding" in display:
                self.settings.set_boolean(
                    ["display_branding"], display["branding"] == 1
                )
            if "while_printing_type" in display:
                wpt = str(display["while_printing_type"])
                self.settings.set(["display_while_printing_type"], wpt)
            if "show_status" in display:
                self.settings.set_boolean(
                    ["display_show_status"], display["show_status"] == 1
                )
        if "has_power_controller" in sp_settings:
            has_pwr = sp_settings["has_power_controller"] == 1
            self.settings.set_boolean(["has_power_controller"], has_pwr)
            if has_pwr:
                self._send_power_state()
        if "has_filament_sensor" in sp_settings:
            has_fs = sp_settings["has_filament_sensor"] == 1
            self.settings.set_boolean(["has_filament_sensor"], has_fs)
            if has_fs:
                self._send_filament_sensor_state()
        # self.settings.set(
        #     ["info", "last_user_settings_sync"],
        #     sp_settings["updated_datetime"]
        # )
        self.settings.save(trigger_event=True)

    def _sync_webcam_settings(self, cam_settings: Dict[str, Any]) -> None:
        self._logger.info(
            "Syncing Webcam Settings at request from SimplyPrint"
        )
        flip_h = cam_settings.get("flipH", False)
        flip_v = cam_settings.get("flipV", False)
        rotate_90 = cam_settings.get("rotate90", False)
        self.settings.global_set(["webcam", "flipH"], flip_h)
        self.settings.global_set(["webcam", "flipV"], flip_v)
        self.settings.global_set(["webcam", "rotate90"], rotate_90)
        data = {"flipH": flip_h, "flipV": flip_v, "rotate90": rotate_90}
        self.settings.set(["webcam"], data)
        self.settings.save()

    def _save_printer_profile(self, sp_profile: Dict[str, Any]) -> None:
        profile_mgr = octoprint.server.printerProfileManager
        current_prof = profile_mgr.get("sp_printer")
        if current_prof is None:
            current_prof = profile_mgr.get_default()
        merged = octoprint.util.dict_merge(current_prof, sp_profile)
        make_default = False
        if "default" in merged:
            make_default = True
            sp_profile.pop("default", None)
            merged["id"] = "sp_printer"
        success = True
        try:
            profile_mgr.save(
                merged, allow_overwrite=True, make_default=make_default,
                trigger_event=True
            )
        except Exception:
            self._logger.exception("Failed to save SimplyPrint Profile")
            success = False
        self._loop.add_callback(
            self.send_sp, "profile_saved", {"success": success}
        )

    def _send_gcode_scripts(self, force: bool = False):
        backed_up = self.settings.get_boolean(
            ["info", "gcode_scripts_backed_up"]
        )
        data: Optional[Dict[str, str]] = None
        if not backed_up or force:
            default_cancel_gc = (
                ";disablemotorsM84;disableallheaters"
                "{%snippet'disable_hotends'%}{%snippet'disable_bed'%};"
                "disablefanM106S0"
            )
            cur_cancel_gc = self.settings.settings.loadScript(
                "gcode", "afterPrintCancelled", source=True
            )

            def remove_whitespace(data: str) -> str:
                return data.strip().replace(" ", "").replace("\n", "")

            if remove_whitespace(cur_cancel_gc) == default_cancel_gc:
                cur_cancel_gc = ""
            cur_resume_gc = self.settings.settings.loadScript(
                "gcode", "beforePrintResumed", source=True
            )
            cur_pause_gc = self.settings.settings.loadScript(
                "gcode", "afterPrintPaused", source=True
            )
            if cur_cancel_gc or cur_pause_gc or cur_resume_gc:
                self._logger.info(
                    "Sending G-Code scripts at request of SimplyPrint"
                )
                data = {
                    "cancel_gcode": cur_cancel_gc.splitlines() if cur_cancel_gc else [],
                    "pause_gcode": cur_pause_gc.splitlines() if cur_pause_gc else [],
                    "resume_gcode": cur_resume_gc.splitlines() if cur_resume_gc else []
                }
            self.settings.set_boolean(["info", "gcode_scripts_backed_up"], True)
            if not cur_cancel_gc.startswith("; synced from SimplyPrint GCODE Macros") or \
                    not cur_resume_gc.startswith("; synced from SimplyPrint GCODE Macros") or \
                    not cur_pause_gc.startswith("; synced from SimplyPrint GCODE Macros"):
                self.send_sp("gcode_scripts", {"scripts": data})

    def _save_gcode_scripts(self, scripts: Dict[str, list]) -> None:
        def fix_script(data: list) -> str:
            data = "\n".join(data)
            return octoprint.util.to_unicode(data)

        if (
            "cancel" in scripts and
            "pause" in scripts and
            "resume" in scripts
        ):
            self._logger.info("Saving G-Code scripts at request of SimplyPrint")
            self.settings.settings.saveScript(
                "gcode", "afterPrintCancelled", fix_script(scripts["cancel"])
            )
            self.settings.settings.saveScript(
                "gcode", "afterPrintPaused", fix_script(scripts["pause"])
            )
            self.settings.settings.saveScript(
                "gcode", "beforePrintResumed", fix_script(scripts["resume"])
            )
            self.settings.save(force=True, trigger_event=True)
            self.send_sp("gcode_scripts", {"saved": True})

    async def _test_webcam(self) -> None:
        await self.webcam_stream.test_connection()
        self.send_sp(
            "webcam_status", {"connected": self.webcam_stream.webcam_connected}
        )

    def _save_item(self, name: str, data: Any):
        self.settings.set([name], data)
        self.settings.save()

    def _set_ws_url(self):
        token: str = self.settings.get(["printer_token"])
        if not token:
            # attempt to fall back to pi id of token is not available
            token = self.settings.get(["rpi_id"])
            if token:
                self._save_item("printer_token", token)
        printer_id: str = self.settings.get(["printer_id"])
        ep = WS_TEST_ENDPOINT if self.test else WS_PROD_ENDPOINT
        self.connect_url = f"{ep}/0/0"
        if token:
            if not printer_id:
                self.connect_url = f"{ep}/0/{token}"
            else:
                self.connect_url = f"{ep}/{printer_id}/{token}"

    def _on_state_event(self, new_state: str) -> None:
        if not self.is_connected:
            return
        self._update_state(new_state)

    def _on_printer_connected(self):
        self.is_connected = True
        self._send_connection_state("connected")
        if self.printer.is_printing():
            self.job_info_timer.start()
            self._update_state("printing")
        elif self.printer.is_operational():
            self._update_state("operational")
            if self.settings.get_boolean(["display_show_status"]):
                self.reset_printer_display_timer.start()
        self.temp_timer.start()
        self.amb_detect.start()
        self.printer_reconnect_timer.stop()
        self._send_active_extruder(0)

    def _on_printer_disconnected(self):
        self.is_connected = False
        self._update_state("offline")
        self._send_connection_state("disconnected")
        self.temp_timer.stop()
        self.job_info_timer.stop()
        self.webcam_stream.stop()
        self.ai_timer.stop()
        self.reset_printer_display_timer.stop()
        self.cache.reset_print_state()
        self._start_printer_reconnect()

    def _on_printer_error(self, msg: str) -> None:
        self._update_state("error")
        self.send_sp("printer_error", {"error": msg})

    def _on_print_start(
        self, print_data: Dict[str, Any], need_start_event: bool = True
    ) -> None:
        # inlcludes started and resumed events
        if self.file_handler.start_pending():
            self.send_sp("file_progress", {"state": "started"})
            self._loop.run_in_executor(None, self.file_handler.notify_started)
        self._update_state("printing")
        filename = print_data["name"]
        dest = print_data["origin"]
        path = print_data["path"]
        metadata: Dict[str, Any] = {}
        if dest != FileDestinations.SDCARD:
            if self.file_manager.has_analysis(dest, path):
                metadata = self.file_manager.get_metadata(dest, path)["analysis"]
        job_info: Dict[str, Any] = {"filename": filename}
        filament: float = 0.
        for tool, data in metadata.get("filament", {}).items():
            if tool.startswith("tool") and "length" in data:
                filament += data["length"]
        if filament:
            job_info["filament"] = round(filament)
        est_time = metadata.get("estimatedPrintTime")
        if est_time is not None:
            job_info["time"] = int(est_time + .5)
        self.cache.job_info.update(job_info)
        if need_start_event:
            job_info["started"] = True
        self.job_info_timer.start()
        self._send_job_event(job_info)
        self.set_display_message("Printing...", True)
        self.ai_timer_not_before = datetime.datetime.now() + datetime.timedelta(seconds=120)
        self.ai_timer.start(delay=120.)
        self.scores = []
        # self.reset_printer_display_timer.stop()

    def _on_print_paused(self) -> None:
        self.job_info_timer.stop()
        self.ai_timer.stop()
        self.send_sp("job_info", {"paused": True})
        self._update_state("paused")
        if self.settings.get_boolean(["display_show_status"]):
            self.set_display_message("Paused")

    def _on_print_resumed(self) -> None:
        self.job_info_timer.start()
        self.ai_timer.start(delay=self.intervals.get("ai"))
        self._update_state("printing")
        self.scores = []

    def _on_print_done(
        self, job_state: str, payload: Optional[Dict[str, Any]] = None
    ) -> None:
        self.job_info_timer.stop()
        self.ai_timer.stop()
        # self.reset_printer_display_timer.start()
        event_payload: Dict[str, Any] = {job_state: True}
        if payload is not None:
            event_payload.update(payload)
        self._send_job_event(event_payload)
        self.cache.job_info = {}
        self.current_layer = -1
        self.set_display_message("Print complete", True)

    def _on_metadata_update(self, payload: Dict[str, Any]) -> None:
        if (
            "result" not in payload or
            payload["result"].get("analysisPending", False)
        ):
            return
        self.file_handler.check_analysis(payload)
        fname = payload.get("name", None)
        fpath = payload.get("path", "")
        if (
            self.printer.is_printing() and
            self.printer.is_current_file(fpath, False) and
            fname == self.cache.job_info.get("filename", "")
        ):
            metadata: Dict[str, Any] = payload["result"]
            job_info: Dict[str, Any] = {}
            filament: float = 0.
            for tool, data in metadata.get("filament", {}).items():
                if tool.startswith("tool") and "length" in data:
                    filament += data["length"]
            if filament:
                job_info["filament"] = round(filament)
            est_time = metadata.get("estimatedPrintTime")
            if est_time is not None:
                job_info["time"] = int(est_time + .5)
            diff = self._get_object_diff(job_info, self.cache.job_info)
            if diff:
                self.cache.job_info.update(diff)
                self.send_sp("job_info", diff)

    def _on_cancelling_event(self, msg: str):
        self.cache.firmware_error = msg
        self._update_state("cancelling")

    async def _handle_cpu_update(self, eventtime: float) -> float:
        sys_stats: Dict[str, Any] = await self._loop.run_in_executor(
            None, self.monitor.get_all_resources
        )
        cpu: float = sys_stats["cpu"]["average"]
        mem_pct: float = sys_stats["memory"].get("percent", 0)
        temp_data: Optional[Dict[str, Any]] = sys_stats["temp"]
        temp: float = 0.
        if isinstance(temp_data, dict):
            temp = temp_data.get("current", 0)
        cpu_data = {
            "usage": int(cpu + .5),
            "temp": int(temp + .5),
            "memory": int(mem_pct + .5),
            "flags": self.cache.throttled_state.get("raw_value", 0)
        }
        diff = self._get_object_diff(cpu_data, self.cache.cpu_info)
        if diff:
            self.cache.cpu_info.update(cpu_data)
            self.send_sp("cpu", diff)
        return eventtime + self.intervals["cpu"]

    def _on_cpu_throttled(self, payload: Dict[str, Any]):
        self.cache.throttled_state = payload

    def _on_ambient_changed(self, new_ambient: int) -> None:
        self._save_item("ambient_temp", new_ambient)
        self.send_sp("ambient", {"new": new_ambient})

    async def _handle_job_info_update(self, eventtime: float) -> float:
        if self.cache.state != "printing":
            return eventtime + self.intervals["job"]
        job_info: Dict[str, Any] = {}
        cur_data = await self._loop.run_in_executor(None, self.printer.get_current_data)
        progress: Dict[str, Any] = cur_data["progress"]
        time_left: Optional[float] = progress.get("printTimeLeft")
        pct_done: Optional[int] = None
        if time_left is not None:
            last_time_left = self.cache.job_info.get("time", time_left + 60.)
            time_diff = last_time_left - time_left
            if (
                (time_left < 60 or time_diff >= 30) and
                time_left != last_time_left
            ):
                job_info["time"] = time_left
                if self.settings.get_int(["display_while_printing_type"]) == 1:
                    remaining_time = str(datetime.timedelta(seconds=time_left))
                    self.set_display_message(f"Time Remaining: {remaining_time}", True)
            if progress.get("PrintTimeLeftOrigin", "") == "genius":
                ptime = progress.get("printTime", 0)
                total = ptime + time_left
                pct_done = int(ptime / total * 100 + .5)
        if pct_done is None and "completion" in progress:
            if progress["completion"] is not None:
                pct_done = int(progress["completion"] + .5)
        if (
            pct_done is not None and
            pct_done != self.cache.job_info.get("progress", 0)
        ):
            job_info["progress"] = pct_done
            if self.settings.get_int(["display_while_printing_type"]) == 0:
                self.set_display_message(f"Printing {pct_done}%", True)
        layer = self.current_layer
        if layer != self.cache.job_info.get("layer", -1):
            job_info["layer"] = layer
        if job_info:
            self.cache.job_info.update(job_info)
            self.send_sp("job_info", job_info)
        return eventtime + self.intervals["job"]

    def _handle_temperature_update(self, eventtime: float) -> float:
        if not self.printer.is_operational():
            return eventtime + self.intervals["temps"]
        current_temps: Dict[str, Any] = self.printer.get_current_temperatures()
        need_rapid_update: bool = False
        temp_data: Dict[str, List[int]] = {}
        for heater, temps in current_temps.items():
            if heater == "chamber":
                continue
            try:
                reported_temp = temps["actual"]
                ret = [
                    int(reported_temp + .5),
                    int(temps["target"] + .5)
                ]
            except Exception:
                continue
            last_temps = self.cache.temps.get(heater, [-100., -100.])
            if ret[1] == last_temps[1]:
                if ret[1]:
                    seeking_target = abs(ret[1] - ret[0]) > 5
                else:
                    seeking_target = ret[0] >= self.amb_detect.ambient + 25
                need_rapid_update |= seeking_target
                # The target hasn't changed and not heating, debounce temp
                if heater in self.last_received_temps and not seeking_target:
                    last_reported = self.last_received_temps[heater]
                    if abs(reported_temp - last_reported) < .75:
                        self.last_received_temps.pop(heater)
                        continue
                if ret[0] == last_temps[0]:
                    self.last_received_temps[heater] = reported_temp
                    continue
                temp_data[heater] = ret[:1]
            else:
                # target has changed, send full data
                temp_data[heater] = ret
            self.last_received_temps[heater] = reported_temp
            self.cache.temps[heater] = ret
        if temp_data:
            self.send_sp("temps", temp_data)
        interval = self.intervals["temps"]
        if need_rapid_update:
            interval = self.intervals["temps_target"]
        return eventtime + interval

    def _start_printer_reconnect(self) -> None:
        reconnect_time = self.intervals.get("reconnect", 0)
        if not reconnect_time or self.printer.is_operational():
            return
        self.printer_reconnect_timer.start(delay=2.)

    async def _handle_printer_reconnect(self, eventtime: float) -> float:
        rt = self.intervals.get("reconnect", 2.)
        if not rt or self.printer.is_operational():
            self.printer_reconnect_timer.stop()
        elif self.printer.get_state_id() != "CONNECTING":
            await self._loop.run_in_executor(None, self.printer.connect)
        return eventtime + rt

    async def _handle_update_check(self, eventtime: float) -> float:
        updates = await self._loop.run_in_executor(
            None, self.sys_manager.check_software_update
        )
        if updates != self.cache.updates:
            self.cache.updates = updates
            self.send_sp("software_updates", {"available": updates})
        return eventtime + UPDATE_CHECK_TIME

    async def _make_ai_request(self, endpoint, data, headers, timeout=10):
        return await self._loop.run_in_executor(
            None,
            functools.partial(
                requests.get, endpoint, data=data, headers=headers,
                timeout=timeout
            )
        )

    async def _post_snapshot(self, id: str = None, timer: float = None,
                             endpoint: str = "https://api.simplyprint.io/jobs/ReceiveSnapshot") -> None:
        if id is not None:
            img_data = await self._loop.run_in_executor(None, self.webcam_stream.extract_image)
            data = {"id": id, "image": img_data}
            headers = {"User-Agent": "Mozilla/5.0"}
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    requests.post, endpoint, data=data, headers=headers,
                    timeout=45
                )
            )
        elif timer is not None:
            self._logger.debug("waiting to take snapshot")
            await asyncio.sleep(timer / 1000)
            img_data = await self._loop.run_in_executor(None, self.webcam_stream.extract_image)
            self._logger.debug("sending image over websocket")
            self.send_sp("stream", {"base": img_data})

    async def _handle_ping(self, eventtime: float) -> float:
        ping_interval = self.intervals.get("ping", 20.)
        self._last_ping_sent = self._monotonic()
        self.send_sp("ping", None)
        return eventtime + ping_interval

    async def _handle_ai_snapshot(self, eventtime: float) -> float:
        ai_interval = self.intervals.get("ai", 0)
        if ai_interval > 0 and self.webcam_stream.webcam_connected and self.printer.is_printing():
            img_data = await self._loop.run_in_executor(None, self.webcam_stream.extract_image)
            headers = {"User-Agent": "Mozilla/5.0"}
            data = json.dumps(
                {
                    "api_key": self.settings.get(["printer_token"]),
                    "image_array": img_data,
                    "interval": ai_interval,
                    "printer_id" : self.settings.get(["printer_id"]),
                    "settings" : {
                        "buffer_percent" : 80,
                        "confidence" : 60,
                        "buffer_length" : 16
                    },
                    "scores" : self.scores
                }
            ).encode('utf8')
            try:
                response = await self._make_ai_request("https://ai.simplyprint.io/api/v2/infer", data=data, headers=headers, timeout=10)
                self.failed_ai_attempt = 0
                response_json = response.json()
                self.scores = response_json.get("scores", self.scores)
                self.send_sp("ai_resp", {"ai" : response_json.get("s1", [0, 0, 0])})
            except:
                self.failed_ai_attempts += 1
                td = ai_interval + (self.failed_ai_attempts * 5.0) if self.failed_ai_attempts <= 10 else 120.
                self.ai_timer.start(delay=td)
                
            self.ai_timer.start(delay=ai_interval)
        else:
            self.ai_timer.stop()
        return eventtime + ai_interval

    def _update_state_from_octo(self) -> None:
        state: str = self.printer.get_state_id()  # type: ignore
        if state in VALID_STATES:
            self._update_state(state.lower())

    def _update_state(self, new_state: str) -> None:
        if self.cache.state == new_state:
            return
        if self.cache.state == "cancelling":
            payload = {"error": self.cache.firmware_error}
            self.cache.firmware_error = ""
            self._on_print_done("cancelled", payload)
        self.cache.state = new_state
        self.send_sp("state_change", {"new": new_state})

    def _send_connection_state(self, conn_state: str) -> None:
        self.send_sp("connection", {"new": conn_state})

    def _send_mesh_data(self, mesh: Dict[str, Any]) -> None:
        self.cache.mesh = mesh
        self.send_sp("mesh_data", mesh)

    def _send_job_event(self, job_info: Dict[str, Any]) -> None:
        if self.connected:
            self.send_sp("job_info", job_info)
        else:
            job_info.update(self.cache.job_info)
            job_info["delay"] = self._monotonic()
            self.missed_job_events.append(job_info)
            if len(self.missed_job_events) > 10:
                self.missed_job_events.pop(0)

    def _on_firmware_warning(self, payload: Dict[str, Any]) -> None:
        self.cache.firmware_warning = payload
        # fw_info = self.cache.firmware_info
        # fw_info["unsafe"] = True
        # self.send_sp("firmware", fw_info)
        self.send_sp("firmware_warning", payload)

    def _send_firmware_data(self, payload: Dict[str, Any]) -> None:
        fw_info = {"fw": payload, "raw": True, "unsafe": False}
        self.cache.firmware_info = fw_info
        self.send_sp("firmware", fw_info)

    def _send_active_extruder(self, new_index: int):
        tool = f"T{new_index}"
        if tool == self.cache.active_extruder:
            return
        self.cache.active_extruder = tool
        self.send_sp("tool", {"new": tool})

    def _send_webcam_config(self) -> None:
        wc_data = {
            "flipH": self.settings.global_get(["webcam", "flipH"]),
            "flipV": self.settings.global_get(["webcam", "flipV"]),
            "rotate90": self.settings.global_get(["webcam", "rotate90"]),
        }
        self.send_sp("webcam", wc_data)

    def _send_image(self, base_image: str) -> None:
        if self._image_delivered:
            self._image_delivered = False
            self.send_sp("stream", {"base": base_image})
        else:
            self._sock_logger.info("Still waiting for last image to be received by server.")

    def _send_installed_plugins(self) -> None:
        sp_plugins = self.settings.get(["sp_installed_plugins"])
        if not isinstance(sp_plugins, list):
            sp_plugins = []
        pm = self.plugin._plugin_manager
        installed_plugins: List[Dict[str, Any]] = []
        plugins: Dict[str, Any] = pm.plugins  # type: ignore
        for plugin in plugins.values():
            if not plugin.bundled and plugin.enabled:
                is_sp = plugin.key.lower() == "simplyprint"
                pinfo = pm.get_plugin_info(plugin.key)  # type: ignore
                if not hasattr(pinfo.origin, "package_name"):
                    # Only report "EntryPointOrigin" packages
                    continue
                installed_plugins.append({
                    "key": plugin.key,
                    "name": plugin.name,
                    "author": plugin.author,
                    "version": plugin.version,
                    "sp_installed": plugin.key in sp_plugins or is_sp,

                    "pip_name": pinfo.origin.package_name
                })
        self.send_sp("installed_plugins", {"plugins": installed_plugins})

    def _send_power_state(self) -> None:
        psu_state = self.sys_manager.get_power_state()
        if psu_state is not None:
            self.send_sp("power_controller", {"on": psu_state})

    def _send_filament_sensor_state(self) -> None:
        fs_state = self.sys_manager.get_filament_sensor_state()
        if fs_state is not None:
            self.send_sp("filament_sensor", {"state": fs_state})

    def _push_initial_state(self):
        # TODO: This method is called after SP is connected
        # and ready to receive state.  We need a list of items
        # we can safely send if the printer is not setup (ie: has no
        # printer ID)
        #
        # The firmware data and machine data is likely saved by
        # simplyprint.  It might be better for SP to request it
        # rather than for the client to send it on every connection.
        self.send_sp("state_change", {"new": self.cache.state})
        if self.cache.temps:
            self.send_sp("temps", self.cache.temps)
        if self.cache.firmware_info:
            self.send_sp("firmware", self.cache.firmware_info)
        if self.cache.machine_data:
            self.send_sp("machine_data", self.cache.machine_data)
        self._send_webcam_config()
        curtime = self._monotonic()
        for evt in self.missed_job_events:
            evt["delay"] = int((curtime - evt["delay"]) + .5)
            self.send_sp("job_info", evt)
        self.missed_job_events = []
        if self.cache.active_extruder:
            self.send_sp("tool", {"new": self.cache.active_extruder})
        if self.cache.cpu_info:
            self.send_sp("cpu_info", self.cache.cpu_info)
        self.send_sp("ambient", {"new": self.amb_detect.ambient})
        self._send_power_state()
        self._send_filament_sensor_state()
        self._send_installed_plugins()
        if self.cache.updates:
            self.send_sp("software_updates", {"available": self.cache.updates})
        self.send_sp(
            "webcam_status", {"connected": self.webcam_stream.webcam_connected}
        )

    def _check_setup_event(self, evt_name: str) -> bool:
        return self.is_set_up or evt_name in PRE_SETUP_EVENTS

    def send_sp(
        self, evt_name: str, data: Any
    ) -> Union[asyncio.Future, asyncio.Task]:
        if (
            not self.connected or
            self.ws is None or
            self.ws.protocol is None or
            not self._check_setup_event(evt_name)
        ):
            fut = self._aioloop.create_future()
            fut.set_result(False)
            return fut
        packet = {"type": evt_name, "data": data}
        if evt_name != "stream":
            self._sock_logger.info(f"sent: {packet}")
        else:
            self._sock_logger.info("sent: webcam stream")
        try:
            fut = self.ws.write_message(json.dumps(packet))
        except tornado.websocket.WebSocketClosedError:
            fut = self._aioloop.create_future()
            fut.set_result(False)
        else:
            async def fut_wrapper():
                try:
                    await fut
                except asyncio.CancelledError:
                    raise
                except Exception:
                    pass
            task = self._aioloop.create_task(fut_wrapper())
            return task
        return fut

    def set_display_message(self, message: str, short_branding=False) -> None:
        enabled = self.settings.get_boolean(["display_enabled"])
        if not self.is_set_up or not enabled:
            return
        if not isinstance(message, str):
            try:
                message = str(message)
            except Exception:
                return
        if message == self.cache.message:
            return
        self.cache.message = message
        if self.settings.get_boolean(["display_branding"]):
            if short_branding or len(message) > 7:
                prefix = "[SP] "
            else:
                prefix = "[SimplyPrint] "
            message = prefix + message
        self._loop.run_in_executor(None, self.printer.commands, f"M117 {message}")

    async def _reset_printer_display(self, eventtime: float) -> float:
        self._logger.debug(f"resetting display at {eventtime}")
        if self.settings.get_boolean(["display_show_status"]) is False:
            self.reset_printer_display_timer.stop()
        is_online = await self._loop.run_in_executor(
            None,
            functools.partial(
                server_reachable, "www.google.com", 80
            ))
        if not is_online:
            self.set_display_message(f"No Internet", True)
        elif self.is_connected and not self.printer.is_printing():
            self.set_display_message(f"Ready")
        return eventtime + self.intervals["ready_message"]

    def _setup_simplyprint_logging(self):
        fpath = self.settings.get_plugin_logfile_path()
        log_path = pathlib.Path(fpath)
        queue: SimpleQueue = SimpleQueue()
        queue_handler = logging.handlers.QueueHandler(queue)
        self._sock_logger = logging.getLogger(
            "octoprint.plugins.simplyprint.test"
        )
        self._sock_logger.addHandler(queue_handler)
        self._sock_logger.propagate = False
        file_hdlr = logging.handlers.TimedRotatingFileHandler(
            log_path, when='midnight', backupCount=2)
        formatter = logging.Formatter(
            '%(asctime)s [%(funcName)s()] - %(message)s')
        file_hdlr.setFormatter(formatter)
        self.qlistner = logging.handlers.QueueListener(queue, file_hdlr)
        self.qlistner.start()

    def _get_object_diff(
        self, new_obj: Dict[str, Any], cached_obj: Dict[str, Any]
    ) -> Dict[str, Any]:
        if not cached_obj:
            return new_obj
        diff: Dict[str, Any] = {}
        for key, val in new_obj.items():
            if key in cached_obj and val == cached_obj[key]:
                continue
            diff[key] = val
        return diff

    async def _do_close(self):
        self._logger.info("Closing SimplyPrint Websocket...")
        self.amb_detect.stop()
        self.temp_timer.stop()
        self.job_info_timer.stop()
        self.cpu_timer.stop()
        self.ai_timer.stop()
        self.ping_timer.stop()
        self.printer_reconnect_timer.stop()
        self.update_timer.stop()
        try:
            await self.send_sp("shutdown", None)
        except tornado.websocket.WebSocketClosedError:
            pass
        self.qlistner.stop()
        self.is_closing = True
        if self.ws is not None:
            self.ws.close(1001, "Client Shutdown")
        if (
            self.connection_task is not None and
            not self.connection_task.done()
        ):
            try:
                await asyncio.wait_for(self.connection_task, 2.)
            except asyncio.TimeoutError:
                pass

    def close(self):
        if self._aioloop.is_running():
            self._loop.add_callback(self._do_close)
            self.simplyprint_thread.join()

    @property
    def event_bus(self):
        return self._event_bus


class ReportCache:
    def __init__(self) -> None:
        self.state = "offline"
        self.temps: Dict[str, Any] = {}
        self.mesh: Dict[str, Any] = {}
        self.job_info: Dict[str, Any] = {}
        self.active_extruder: str = ""
        # Persistent state across connections
        self.firmware_info: Dict[str, Any] = {}
        self.firmware_warning: Dict[str, Any] ={}
        self.machine_data: Dict[str, Any] = {}
        self.cpu_info: Dict[str, Any] = {}
        self.throttled_state: Dict[str, Any] = {}
        self.download_progress: int = -1
        self.message: str = ""
        self.updates: List[Dict[str, Any]] = []
        self.firmware_error: str = ""

    def reset_print_state(self) -> None:
        self.temps = {}
        self.mesh = {}
        self.job_info = {}

AMBIENT_CHECK_TIME = 5. * 60.
TARGET_CHECK_TIME = 60. * 60.
SAMPLE_CHECK_TIME = 20.

class AmbientDetect:
    CHECK_INTERVAL = 5
    def __init__(
        self,
        cache: ReportCache,
        changed_cb: Callable[[int], None],
        initial_ambient: int
    ) -> None:
        self._logger = logging.getLogger(
            "octoprint.plugins.simplyprint")
        self.cache = cache
        self._initial_sample: int = -1000
        self._ambient = initial_ambient
        self._on_ambient_changed = changed_cb
        self._last_sample_time: float = 0.
        self._update_interval = AMBIENT_CHECK_TIME
        self._detect_timer = FlexTimer(self._handle_detect_timer)

    @property
    def ambient(self) -> int:
        return self._ambient

    def _handle_detect_timer(self, eventtime: float) -> float:
        if "tool0" not in self.cache.temps:
            self._initial_sample = -1000
            return eventtime + self.CHECK_INTERVAL
        temp, target = self.cache.temps["tool0"]
        if target:
            self._initial_sample = -1000
            self._last_sample_time = eventtime
            self._update_interval = TARGET_CHECK_TIME
            return eventtime + self.CHECK_INTERVAL
        if eventtime - self._last_sample_time < self._update_interval:
            return eventtime + self.CHECK_INTERVAL
        if self._initial_sample == -1000:
            self._initial_sample = temp
            self._update_interval = SAMPLE_CHECK_TIME
        else:
            diff = abs(temp - self._initial_sample)
            if diff <= 2:
                last_ambient = self._ambient
                self._ambient = int((temp + self._initial_sample) / 2 + .5)
                self._initial_sample = -1000
                self._last_sample_time = eventtime
                self._update_interval = AMBIENT_CHECK_TIME
                if last_ambient != self._ambient:
                    self._logger.debug(f"SimplyPrint: New Ambient: {self._ambient}")
                    self._on_ambient_changed(self._ambient)
            else:
                self._initial_sample = temp
                self._update_interval = SAMPLE_CHECK_TIME
        return eventtime + self.CHECK_INTERVAL

    def start(self) -> None:
        if self._detect_timer.is_running():
            return
        if "tool0" in self.cache.temps:
            cur_temp = self.cache.temps["tool0"][0]
            if cur_temp < self._ambient:
                self._ambient = cur_temp
                self._on_ambient_changed(self._ambient)
        self._detect_timer.start()

    def stop(self) -> None:
        self._detect_timer.stop()

class FlexTimer:
    def __init__(self, callback: TimerCallback) -> None:
        self.callback = callback
        self.timer_handle: Optional[asyncio.Handle] = None
        self.running: bool = False

    def start(self, delay: float = 0.):
        if self.running:
            return
        if not hasattr(self, "_aioloop"):
            self._aioloop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self.running = True
        call_time = self._aioloop.time() + delay
        self.timer_handle = self._aioloop.call_at(
            call_time, self._schedule_task
        )

    def stop(self):
        if not self.running:
            return
        self.running = False
        if self.timer_handle is not None:
            self.timer_handle.cancel()
            self.timer_handle = None

    def _schedule_task(self):
        self.timer_handle = None
        self._aioloop.create_task(self._call_wrapper())

    def is_running(self) -> bool:
        return self.running

    async def _call_wrapper(self):
        if not self.running:
            return
        ret = self.callback(self._aioloop.time())
        if isinstance(ret, Awaitable):
            ret = await ret
        if self.running:
            self.timer_handle = self._aioloop.call_at(ret, self._schedule_task)



import psutil
import time


class Monitor:

	def __init__(self, logger):
		self.__logger = logger
		self.__init_process()
		self.__init_children()

	def __get_cpu_temp(self, temp):
		if temp is None:
			return None
		cpu_temp = None
		if "coretemp" in temp:
			cpu_temp = temp["coretemp"][0]._asdict()
		if "cpu-thermal" in temp:
			cpu_temp = temp["cpu-thermal"][0]._asdict()
		if "cpu_thermal" in temp:
			cpu_temp = temp["cpu_thermal"][0]._asdict()
		if "soc_thermal" in temp:
			cpu_temp = temp["soc_thermal"][0]._asdict()
		return cpu_temp

	def __init_process(self):
		self.__process = psutil.Process()
		self.__logger.debug("self.__process is now %r" % (self.__process,))
		# First call so it does not return 0 on next call
		self.__process.cpu_percent()

	def get_cpu(self):
		cpu_freq = psutil.cpu_freq()
		self.__logger.debug("cpu_freq() : %r" % (cpu_freq,))
		cores = psutil.cpu_percent(percpu=True)
		self.__logger.debug("cpu_percent(percpu=True) : %r" % (cores,))
		average = psutil.cpu_percent()
		self.__logger.debug("cpu_percent() : %r" % (average,))
		core_count = psutil.cpu_count(logical=False)
		self.__logger.debug("cpu_count(logical=False) : %r" % (core_count,))
		thread_count = psutil.cpu_count(logical=True)
		self.__logger.debug("cpu_count(logical=True) : %r" % (thread_count,))
		pids = len(psutil.pids())
		self.__logger.debug("len(pids()) : %r" % (pids,))
		boot_time = psutil.boot_time()
		self.__logger.debug("boot_time() : %r" % (boot_time,))
		return dict(
			cores=cores,
			average=average,
			frequency=cpu_freq._asdict() if cpu_freq else dict(),
			core_count=core_count,
			thread_count=thread_count,
			pids=pids,
			uptime=int(time.time() - boot_time),
			octoprint=self.__get_octoprint_cpu(average)
		)

	def __init_children(self):
		try:
			self.__children = self.__process.children(recursive=True)
		except psutil.NoSuchProcess:
			self.__logger.debug("No process found when calling children(recursive=True) on %r" % (self.__process,))
			self.__init_process()
			self.__children = self.__process.children(recursive=True)
		for child in self.__children:
			# First call so it does not return 0 on next call, process might die between calls
			try:
				child.cpu_percent()
			except psutil.NoSuchProcess:
				self.__logger.debug("No process found when calling cpu_percent() on %r" % (child,))
				pass

	def __get_octoprint_cpu(self, average):
		try:
			# Don't know why and how
			# But his can sometimes raise a NoSuchProcessException
			total_cpu = self.__process.cpu_percent()
		except psutil.NoSuchProcess:
			self.__logger.debug("No process found when calling cpu_percent() on %r" % (self.__process,))
			self.__init_process()
			total_cpu = self.__process.cpu_percent()
		for child in self.__children:
			try:
				total_cpu += child.cpu_percent()
			except psutil.NoSuchProcess:
				self.__logger.debug("No process found when calling cpu_percent() on %r" % (child,))
				pass
		self.__init_children()
		cpu_count = psutil.cpu_count()
		self.__logger.debug("cpu_count() : %r" % (cpu_count,))
		return min(total_cpu / cpu_count, average)

	def get_cpu_temp(self):
		# return dict(
		#   current=20 (celsius only)
		# )
		temps_celsius = None
		if hasattr(psutil, "sensors_temperatures"):
			temps_celsius = psutil.sensors_temperatures()
			self.__logger.debug("sensors_temperatures() : %r" % (temps_celsius,))
		temps_c = self.__get_cpu_temp(temps_celsius)
		return temps_c

	def get_memory(self):
		virtual_memory = psutil.virtual_memory()
		self.__logger.debug("virtual_memory() : %r" % (virtual_memory,))
		return virtual_memory._asdict()

	def get_all_resources(self):
		return dict(
			cpu=self.get_cpu(),
			temp=self.get_cpu_temp(),
			memory=self.get_memory(),
		)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import socket

from octoprint.settings import settings
import octoprint.server

try:
    # Python 3
    from urllib.parse import quote
except ImportError:
    # Python 2
    from urllib import quote


def url_quote(string):
    return quote(string)


def has_internet():
    try:
        socket.create_connection(("www.google.com", 80))
        return True
    except OSError:
        return False


def is_octoprint_setup():
    return not (
        settings().getBoolean(["server", "firstRun"]) and (
            octoprint.server.userManager is None
            or not octoprint.server.userManager.has_been_customized()
        )
    )


def any_demand(demand_list, demands):
    for demand in demands:
        if demand in demand_list:
            return True

    return False


# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse


def run_script(name):
    # Currently there's only one script, but it could change :)
    if name == "run_healthcheck":
        from octoprint_simplyprint.local.background import run_background_check
        run_background_check()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="SimplyPrint Local Scripts")
    parser.add_argument("script",
                        default=None,
                        help="The script you want to call, currently only `run_healthcheck` is valid")
    args = parser.parse_args()
    run_script(args.script)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, unicode_literals
#
# SimplyPrint
# Copyright (C) 2020-2021  SimplyPrint ApS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import sys
import subprocess

from crontab import CronTab

background_comment = "[SimplyPrint] Background healthcheck (V1.1)"
background_command = sys.executable + " -m octoprint_simplyprint run_healthcheck"


class CronManager:
    def __init__(self):
        self.cron = CronTab(user=True)
        self._logger = logging.getLogger()

        for job in self.cron:
            comment = job.comment.lower()
            if (
                ("simplyprint" in comment and comment != background_comment.lower())
                # Remove commands with the simplyrint in them, this was a typo in previous versions
                or ("simplyrint" in job.command and comment != background_comment.lower())
            ):
                self._logger.info("Removing job, {}".format(comment))
                self.cron.remove(job)

    def add(self, user, command, comment, on_reboot=False, daily=False):
        exist_check = self.cron.find_comment(comment)
        try:
            if sys.version_info.major == 3:
                job = next(exist_check)
            else:
                job = exist_check.next()

            if len(job) > 0:
                self._logger.info("Cronjob for SimplyPrint check already exists - not creating")
                return True

        except StopIteration:
            pass

        self._logger.info("Creating cronjob...")

        cron_job = self.cron.new(command=command, comment=comment, user=user)

        if on_reboot:
            cron_job.every_reboot()
        else:
            if not daily:
                cron_job.minute.every(1)
            else:
                cron_job.hour.on(0)
                cron_job.minute.on(0)

        cron_job.enable()
        self.cron.write()
        return True

    def validate(self, comment):
        """
        Validates that an entry with the comment exists in the crontab
        """
        for job in self.cron:
            job_comment = job.comment.lower()
            if "simplyprint" in job_comment and job_comment == comment.lower():
                return True
        return False

    def remove(self, comment):
        for job in self.cron:
            if comment.lower() == job.comment.lower():
                self.cron.remove(job)


def create_cron_jobs():
    cron = CronManager()
    cron.add(True, background_command, background_comment, True)

    # Start the background process *now*, but don't wait for it (it will run forever...)
    subprocess.Popen([sys.executable, "-m", "octoprint_simplyprint", "run_healthcheck"])


def check_cron_jobs():
    cron = CronManager()
    return cron.validate(background_comment)


def remove_cron_jobs():
    cron = CronManager()
    cron.remove(background_comment)
